"""
Measure T1 and T2 in Caliber relaxometry phantom.

Introduction
============

This module determines the T1 and T2 decay constants for the relaxometry
spheres in the Caliber (HPD) system phantom
qmri.com/qmri-solutions/t1-t2-pd-imaging-phantom (plates 4 and 5).
Values are compared to published values (without temperature correction).
Graphs of fit and phantom registration images can optionally be produced.


Scan parameters
===============

Manufacturer's details of recommended scan parameters for GE, Philips and
Siemens scanners are available in the 'System Phantom Manual' which can be
downloaded from the above website (T1-VTI and T2 sequences). However, these may
result in long scan times. The parameters below were used to acquire the
images used in testing this module. They are provided for information only.


T1 Relaxometry
--------------

Sequence: Spin echo with inversion recovery
Plane: Coronal
TR (ms): 1000 (or minimum achievable if longer--see note)
TE (ms): 10
TI (ms): {50.0, 100.0, 200.0, 400.0, 600.0, 800.0}
Flip angle: 180 degrees
Matrix: 192 x 192
FoV (mm): 250 x 250
Slices: 2 (or 3 to acquire plate 3 with PD spheres)
Slice width (mm): 5
Distance factor: 35 mm / 700%
NSA: 1
Receive bandwidth:
    GE (kHz): 15.63
    Philips (Hz / px): 109
    Siemens (Hz / px): 130
Reconstruction: Normalised

Note: Some scanners may require a longer TR for long TI values. This algorithm
will accommodate a variation in TR with TI and incomplete recovery due to short
TR.


T2 Relaxometry
--------------

Sequence:
    GE: T2 map (TE values fixed)
    Other manufacturers: Spin echo multi contrast
Plane: Coronal
TR (ms): 2000
Number of contrasts: maximum
TE (ms): minimum
Flip angle: 90 degrees
Matrix: 192 x 192
FoV (mm): 250 x 250
Slices: 2 (or 3 to acquire plate 3 with PD spheres)
Slice width (mm): 5
Distance factor: 35 mm / 700%
NSA: 1
Receive bandwidth:
    GE (kHz): 15.63
    Philips (Hz / px): 109
    Siemens (Hz / px): 130
Reconstruction: Normalised


Algorithm overview
==================
1. Create ``T1ImageStack`` or ``T2ImageStack`` object which stores a list
    of individual DICOM files (as ``pydicom`` objects) in the ``.images``
    attribute.
2. Obtain the RT (rotation / translation) matrix to register the template
    image to the test image. Four template images are provided, one for
    each relaxation parameter (T1 or T2) on plates 4 and 5, and regression
    is performed on the first image in the sequence. Optionally output the
    overlay image to visually check the fit.
3. A ROI is generated for each target sphere using stored coordinates, the RT
    transformation above, and a structuring element (default is a 5x5 boxcar).
4. Store pixel data for each ROI at various times, in an ``ROITimeSeries``
    object. A list of these objects is stored in
    ``ImageStack.ROI_time_series``.
5. Generate the fit function. For T1 this looks up TR for the given TI
    (using piecewise linear interpolation if required) and determines if a
    magnitude or signed image is used. No customisation is required for T2
    measurements.
6. Determine relaxation time (T1 or T2) by fitting the decay equation to
    the ROI data for each sphere. The published values of the relaxation
    times are used to seed the optimisation algorithm. A Rician noise model is
    used for T2 fitting [1]_. Optionally plot and save the decay curves.
7. Return plate number, relaxation type (T1 or T2), measured relaxation
    times, published relaxation times, and fractional differences in a
    dictionary.

References
==========
.. [1] Raya, J.G., Dietrich, O., Horng, A., Weber, J., Reiser, M.F.
and Glaser, C., 2010. T2 measurement in articular cartilage: impact of the
fitting method on accuracy and precision at low SNR. Magnetic Resonance in
Medicine: An Official Journal of the International Society for Magnetic
Resonance in Medicine, 63(1), pp.181-193.

Feature enhancements
====================
Template fit on bolt holes--possibly better with large rotation angles
    -have bolthole template, find 3 positions in template and image, figure out
    transformation.

Template fit on outline image--possibly run though edge detection algorithms
then fit.

Use normalised structuring element in ROITimeSeries. This will allow correct
calculation of mean if elements are not 0 or 1.

Get r-squared measure of fit.

"""

import json
import os.path
import pathlib

import cv2 as cv
import hazenlib.exceptions
import matplotlib.pyplot as plt
import numpy as np
import pydicom
import scipy.ndimage
import scipy.optimize
import skimage.morphology
from hazenlib.data.relaxometry_params import (
    MAX_RICIAN_NOISE,
    SEED_RICIAN_NOISE,
    SMOOTH_TIMES,
    TEMPLATE_FIT_ITERS,
    TEMPLATE_VALUES,
    TERMINATION_EPS,
)
from hazenlib.HazenTask import HazenTask
from hazenlib.logger import logger
from hazenlib.types import Measurement, Metadata
from hazenlib.utils import dcmread
from scipy.interpolate import UnivariateSpline
from scipy.special import i0e, ive

# Use dict to store template and reference information
# Coordinates are in array format (row,col), rather than plt.patches
# format (col,row)
#
# Access as:
#    TEMPLATE_VALUES[f'plate{plate_num}']['sphere_centres_row_col']
#    TEMPLATE_VALUES[f'plate{plate_num}']['t1'|'t2']['filename']
#    TEMPLATE_VALUES[f'plate{plate_num}']['t1'|'t2']['1.5T'|'3.0T']['relax_times']


class Relaxometry(HazenTask):
    def __init__(self, **kwargs):
        self.calc = kwargs.pop("calc", "T1")
        self.plate_number = kwargs.pop("plate_number", None)
        super().__init__(**kwargs)

    def run(
        self,
        calc: str | None = None,
        plate_number: int | None = None,
        verbose: bool = False,
    ):
        """
        Calculate T1 or T2 values for relaxometry phantom.

        Note: either ``calc_t1`` or ``calc_t2`` (but not both) must be True.
        Variables set in parent class:
        data : list of pydicom.dataset.FileDataSet objects
            List of DICOM images of a plate of the HPD relaxometry phantom.
        report : bool, optional
            Whether to save images showing the measurement details
        report_dir : path, optional
            Folder path to save images to. The default is False.

        Parameters
        ----------
        calc : str, required
            Whether to calculate T1 or T2 relaxation. Default is T1.
        plate_number : str, required
            Plate number of the HPD relaxometry phantom (either 4 or 5)
        verbose : bool, optional
            Provide verbose output. If True, the following key / values will be
            added to the output dictionary:
                plate : plate_number,
                relaxation_type : 't1' | 't2',
                calc_times : list of T1|T2 for each sphere,
                manufacturers_times : list of manufacturer's values for T1|T2,
                frac_time_difference : (calc_times - manufacturers_times) / manufacturers_times
                institution_name=index_im.InstitutionName,
                manufacturer=index_im.Manufacturer,
                model=index_im.ManufacturerModelName,
                date=index_im.StudyDate,
                detailed_output : dict with extensive information
            The default is False.

        Returns
        -------

        dict
            {
                rms_frac_time_difference : RMS fractional difference between
                    calculated relaxometry times and manufacturer provided data.
            }
        """
        calc = self.calc if calc is None else calc
        plate_number = (
            self.plate_number if plate_number is None else plate_number
        )

        # check plate number specified: should be either 4 or 5
        try:
            plate_number = int(plate_number)  # convert to int if required
        except (ValueError, TypeError):
            pass  # will raise error at next statement
        if plate_number not in [4, 5]:
            raise hazenlib.exceptions.ArgumentCombinationError(
                "Must specify plate_number (4 or 5)"
            )

        # Set up parameters specific to T1 or T2
        relax_str = calc.lower()

        if calc in ["T1", "t1"]:
            image_stack = T1ImageStack(self.dcm_list)
            try:
                template_dcm = dcmread(
                    TEMPLATE_VALUES[f"plate{plate_number}"][relax_str][
                        "filename"
                    ]
                )
            except KeyError:
                logger.exception(
                    "Could not find template with plate number: %s."
                    " Please pass plate number as arg.",
                    plate_number,
                )
                exit()
        elif calc in ["T2", "t2"]:
            image_stack = T2ImageStack(self.dcm_list)
            try:
                template_dcm = dcmread(
                    TEMPLATE_VALUES[f"plate{plate_number}"][relax_str][
                        "filename"
                    ]
                )
            except KeyError:
                logger.exception(
                    "Could not find template with plate number: %s."
                    " Please pass plate number as arg.",
                    plate_number,
                )
                exit()
        else:
            logger.critical(
                "Please provide 'T1' or 'T2' for the --calc argument.",
            )
            exit()

        warp_matrix = image_stack.template_fit(template_dcm)
        image_stack.generate_time_series(
            TEMPLATE_VALUES[f"plate{plate_number}"]["sphere_centres_row_col"],
            warp_matrix=warp_matrix,
        )
        # only applies to T1
        image_stack.generate_fit_function()

        # Published relaxation time for matching plate and T1/T2
        relax_published = TEMPLATE_VALUES[f"plate{plate_number}"][relax_str][
            "relax_times"
        ][image_stack.b0_str]
        s0_est = image_stack.initialise_fit_parameters(relax_published)

        image_stack.find_relax_times(relax_published, s0_est)
        frac_time_diff = (
            image_stack.relax_times - relax_published
        ) / relax_published
        # last value is for background water. Strip before calculating RMS frac error
        frac_time = frac_time_diff[:-1]
        RMS_frac_error = np.sqrt(np.mean(np.square(frac_time)))

        # Generate results dict
        index_im = self.dcm_list[0]
        results = self.init_result_dict()
        output_key = "_".join(
            [self.img_desc(index_im), str(plate_number), relax_str]
        )
        results.files = output_key

        results.add_measurement(
            Measurement(
                name="Relaxometry",
                type="measured",
                subtype="rms_frac_time_difference",
                value=round(RMS_frac_error, 3),
            ),
        )

        if self.report:
            img_path = os.path.join(self.report_path, output_key)
            # Show template fit
            template_fit_fig = image_stack.plot_fit()
            # Improve saved image quality
            template_fit_fig.set_size_inches(24, 24)
            for subplt in template_fit_fig.get_axes():
                subplt.title.set_fontsize(40)
            template_fit_img = f"{img_path}_template_fit.png"
            template_fit_fig.savefig(template_fit_img, dpi=150)
            self.report_files.append(("template_fit", template_fit_img))

            # Show ROIs
            roi_fig = image_stack.plot_rois()
            plt.title(f"ROI positions ({calc.upper()}, plate {plate_number})")
            roi_img = f"{img_path}_rois.png"
            roi_fig.savefig(roi_img, dpi=300)
            self.report_files.append(("rois", roi_img))

            # Show relax fits
            rois = image_stack.ROI_time_series
            relax_fit_fig = plt.figure()
            relax_fit_fig.suptitle(calc.upper() + " relaxometry fits")
            smooth_times = SMOOTH_TIMES[calc.lower()]
            for i in range(15):
                plt.subplot(5, 3, i + 1)
                plt.plot(
                    smooth_times,
                    image_stack.fit_function(
                        np.array(smooth_times),
                        *np.array(image_stack.relax_fit[i][0]),
                    ),
                    "b-",
                )
                plt.plot(rois[i].times, rois[i].means, "rx")
                if i == 14:
                    plt.title(
                        f"[Free water] fit={image_stack.relax_times[i]:.4g}",
                        fontsize=8,
                    )
                else:
                    plt.title(
                        f"[{i + 1}] fit={image_stack.relax_times[i]:.4g}, "
                        f"pub={relax_published[i]:.4g} "
                        f"({frac_time_diff[i] * 100:+.2f}%)",
                        fontsize=8,
                    )
            # Improve saved image quality
            relax_fit_fig.set_size_inches(9, 15)
            plt.tight_layout(rect=(0, 0, 1, 0.97))
            relax_fit_img = f"{img_path}_decay_graphs.png"
            relax_fit_fig.savefig(relax_fit_img, dpi=300)
            self.report_files.append(("decay_graphs", relax_fit_img))

        if verbose:
            # Dump additional details about the images and the measurement to a file
            pathlib.Path(self.report_path).mkdir(parents=True, exist_ok=True)
            detailed_output = {}
            detailed_outpath = os.path.join(
                self.report_path, f"{output_key}_details.json"
            )

            metadata = Metadata(
                files=[im.filename for im in image_stack.images],
                institution_name=index_im.InstitutionName,
                manufacturer=index_im.Manufacturer,
                model=index_im.ManufacturerModelName,
                date=index_im.StudyDate,
            )
            results.metadata = metadata

            for idx, (
                measured_time,
                manufacturers_time,
                frac_time_difference,
            ) in enumerate(
                zip(
                    image_stack.relax_times,
                    relax_published.tolist(),
                    frac_time_diff.tolist(),
                    strict=True,
                ),
            ):
                results.add_measurement(
                    Measurement(
                        name="Relaxometry",
                        value=measured_time,
                        type="measured",
                        subtype=(
                            f"Plate {plate_number}"
                            f" {calc.upper()}"
                            f" Measurement {str(idx).zfill(2)}"
                        ),
                        description="Measured relaxation time.",
                        unit="ms",
                        visibility="intermediate",
                    ),
                )
                results.add_measurement(
                    Measurement(
                        name="Relaxometry",
                        value=manufacturers_time,
                        type="theoretical",
                        subtype=(
                            f"Plate {plate_number}"
                            f" {calc.upper()}"
                            f" Measurement {str(idx).zfill(2)}"
                        ),
                        description="Manufacturer's relaxation time.",
                        unit="ms",
                        visibility="intermediate",
                    ),
                )
                results.add_measurement(
                    Measurement(
                        name="Relaxometry",
                        value=frac_time_difference,
                        type="measured",
                        subtype=(
                            f"Plate {plate_number}"
                            f" {calc.upper()}"
                            f" Measurement {str(idx).zfill(2)}"
                        ),
                        description="Fractional Time Difference.",
                        unit="",
                        visibility="intermediate",
                    ),
                )

            detailed_output["measurement details"] = {
                "Echo Time": [im.EchoTime for im in image_stack.images],
                "Repetition Time": [
                    im.RepetitionTime for im in image_stack.images
                ],
                "Inversion Time": [
                    im.InversionTime if hasattr(im, "InversionTime") else None
                    for im in image_stack.images
                ],
                # fit_paramters (T1) = [[T1, s0, A1] for each ROI]
                # fit_parameters (T2) = [[T2, s0, C] for each ROI]
                "ROI_means": {
                    i: im.means
                    for i, im in enumerate(image_stack.ROI_time_series)
                },
                "fit_parameters": [
                    tuple(param[0].tolist()) for param in image_stack.relax_fit
                ],
                "fit_equation": image_stack.fit_eqn_str,
            }
            detailed_output["metadata"] = metadata.to_dict()
            json_object = json.dumps(detailed_output, indent=4)
            with open(detailed_outpath, "w") as f:
                f.write(json_object)
            self.report_files.append(("further_details", detailed_outpath))

        if self.report:
            results.add_report_image(self.report_files)

        # plt.show()
        return results


def outline_mask(im):
    """
    Create contour lines to outline pixels.

    Creates a series of ``line`` objects to outline contours on an image. Used
    to add ROIs from a mask array. Adapted from [1]_

    Parameters
    ----------
    im : array
        Pixel array used to create outlines. Array values should be 0 or 1.

    Returns
    -------
    lines : list
        List of coordinates of outlines (see Example below).

    Example
    -------
    >>> lines = outline_mask(combined_ROI_map)
    >>> for line in lines:
            plt.plot(line[1], line[0], color='r', alpha=1)

    References
    ----------
    .. [1] stackoverflow.com/questions/40892203/can-matplotlib-contours-match-pixel-edges

    """
    lines = []
    pad = np.pad(im, [(1, 1), (1, 1)])  # zero padding

    im0 = np.abs(np.diff(pad, n=1, axis=0))[:, 1:]
    im1 = np.abs(np.diff(pad, n=1, axis=1))[1:, :]

    im0 = np.diff(im0, n=1, axis=1)
    starts = np.argwhere(im0 == 1)
    ends = np.argwhere(im0 == -1)
    lines += [
        ([s[0] - 0.5, s[0] - 0.5], [s[1] + 0.5, e[1] + 0.5])
        for s, e in zip(starts, ends)
    ]

    im1 = np.diff(im1, n=1, axis=0).T
    starts = np.argwhere(im1 == 1)
    ends = np.argwhere(im1 == -1)
    lines += [
        ([s[1] + 0.5, e[1] + 0.5], [s[0] - 0.5, s[0] - 0.5])
        for s, e in zip(starts, ends)
    ]

    return lines


def transform_coords(
    coords, rt_matrix, input_row_col=True, output_row_col=True
):
    """
    Convert coordinates using RT transformation matrix.

    Note that arrays containing pixel information as displayed using
    plt.imshow(pixel_array), for example are referenced using the row_col (y,x)
    notation, e.g. pixel_array[row,col]. Plotting points or patches using
    matplotlib requires col_row (x,y) notation, e.g. plt.scatter(col,row). The
    correct input and output notation must be selected for the correct
    transformation.

    Parameters
    ----------
    coords : np.array or tuple
        Array (n,2) of coordinates to transform.
    rt_matrix : np.array
        Array (2,3) of transform matrix (Rotation and Translation). See e.g.
        cv2.transform() for details.
    input_row_col : bool, optional
        Select the input coordinate format relative to the image.
        If True, input array has row (y-coordinate) first, i.e.:
            [[row_1,col_1],
             [row_2,col_2],
             ...,
             [row_n,col_n]].
        If False, input array has col (x-coordinate) first, i.e.:
            [[col_1,row_1],
             [col_2,row_2],
             ...,
             [col_n,row_n].
        The default is True.
    output_row_col : bool, optional
        Select the output coordinate order. If True, output matrix is in
        row_col order, otherwise it is in col_row order. The default is True.

    Returns
    -------
    out_coords : np.array
        Returns (n,2) array of transformed coordinates.

    """
    in_coords = np.array(coords)  # ensure using np array

    if input_row_col:  # convert to col_row (xy) format
        in_coords = np.flip(in_coords, axis=1)

    out_coords = cv.transform(np.array([in_coords]), rt_matrix)[
        0
    ]  # reduce to two dimensions
    # out_coords = out_coords

    if output_row_col:
        out_coords = np.flip(out_coords, axis=1)

    return out_coords


def pixel_rescale(dcm):
    """
    Transforms pixel values according to scale values in DICOM header.

    DICOM pixel values arrays cannot directly represent signed or float values.
    This function converts the ``.pixel_array`` using the scaling values in the
    DICOM header.

    For Philips scanners the private DICOM fields 2005,100d (=SI) and 2005,100e
    (=SS) are used as inverse scaling factors to perform the inverse
    transformation [1]_.

    Parameters
    ----------
    dcm : Pydicom.dataset.FileDataset
        DICOM file containing one image.

    Returns
    -------
    numpy.array
        Values in ``dcm.pixel_array`` transformed using DICOM scaling.

    References
    ----------
    .. [1] Chenevert, Thomas L., et al. "Errors in quantitative image analysis
     due to platform-dependent image scaling." Translational Oncology 7.1
     (2014): 65-71. https://www.ncbi.nlm.nih.gov/pmc/articles/PMC3998685/

    """
    # Check for Philips
    if dcm.Manufacturer.startswith("Philips"):
        ss = dcm["2005100e"].value  # Scale slope
        si = dcm["2005100d"].value  # Scale intercept

        return (dcm.pixel_array - si) / ss
    else:
        return pydicom.pixels.apply_modality_lut(dcm.pixel_array, dcm)


class ROITimeSeries:
    """
    Samples at one image location (ROI) at numerous sample times.

    Estimating T1 and T2 relaxation parameters at any ROI requires a series
    of pixel values and sequence times (e.g. TI, TE, TR). This class is a
    wrapper for storing and accessing these parameters.

    Attributes
    ----------
    POI_mask : array
        Array the same size as the image. All values are 0, except a single 1
        at the point of interest (POI), the centre of the ROI.

    ROI_mask : array
        Array the same size as the image. Values in the ROI are coded as 1s,
        all other values are zero.

    pixel_values : list of arrays
        List of 1-D arrays of pixel values in ROI. The variance could be used
        as a measure of ROI homogeneity to identify  incorrect sphere location.

    times : list of floats
        If ``time_attr`` was used in the constructor, this list contains the
        value of ``time_attr``. Typically ``'EchoTime'`` or
        ``'InversionTime'``.

    trs : list of floats
        Values of TR for each image.

    means : list of floats
        Mean pixel value of ROI for each image in series.
    """

    def __init__(self, dcm_images, poi_coords_row_col, time_attr):
        """
        Create ROITimeSeries for ROI parameters at sequential scans.

        Parameters
        ----------
        dcm_images : list
            List of pydicom images of same object with different scan
            parameters (e.g. TIs or TEs). Typically ``ImageStack.images``.
        poi_coords_row_col : array
            Two element array with coordinates of point of interest (POI),
            typically the centre of the ROI, in row_col (y,x) format.
        time_attr : string, optional
            If present, lookup the DICOM attribute ``[time_attr]`` (typically
            ``'InversionTime'`` or ``'EchoTime'``) and store in the list
            ``self.times``. The default is ``None``, which does not create
            ``self.times``
        """

        self.POI_mask = np.zeros(
            (
                dcm_images[0].pixel_array.shape[0],
                dcm_images[0].pixel_array.shape[1],
            ),
            dtype=np.int8,
        )
        self.POI_mask[poi_coords_row_col[0], poi_coords_row_col[1]] = 1

        kernel = skimage.morphology.square(5)
        """kernel
            Structuring element which defines ROI size and shape, centred on
            POI. Each element should be 1 or 0, otherwise calculation of mean
            will be incorrect. If ``None``, use a 5x5 square. The default is
            ``None``.
        """

        self.ROI_mask = np.zeros_like(self.POI_mask)
        self.ROI_mask = scipy.ndimage.filters.convolve(self.POI_mask, kernel)

        self.times = [x[time_attr].value.real for x in dcm_images]
        self.pixel_values = [
            pixel_rescale(img)[self.ROI_mask > 0] for img in dcm_images
        ]

        self.trs = [x["RepetitionTime"].value.real for x in dcm_images]

    def __len__(self):
        """Number of time samples in series."""
        return len(self.pixel_values)

    @property
    def means(self):
        """
        List of mean ROI values at different times.

        Returns
        -------
        List of mean pixel value in ROI for each sample.
        """
        return [np.mean(pvs) for pvs in self.pixel_values]


class ImageStack:
    """
    Object to hold image_slices and methods for T1, T2 calculation.
    """

    def __init__(self, image_slices, time_attribute=None):
        """
        Create ImageStack object.

        Parameters
        ----------
        image_slices : list of pydicom.FileDataSet objects
            List of pydicom objects to perform relaxometry analysis on.

        template_dcm : pydicom FileDataSet (or None)
            DICOM template object.

        plate_number : int {3,4,5}, optional
            For future use. Reference to the plate in the relaxometry phantom.
            The default is None.

        time_attribute : string, optional
            DICOM attribute to order images. Typically 'InversionTime' for T1
            relaxometry or 'EchoTime' for T2.
        """
        self.time_attr = time_attribute
        # store sorted images
        self.images = self.order_by(image_slices, time_attribute)

        b0_val = self.images[0]["MagneticFieldStrength"].value
        if b0_val not in [1.5, 3.0]:
            # TODO incorporate warning through e.g. logging module
            logger.warning(
                "Unable to match B0 to default values. Using 1.5T.\n%s",
                self.images[0]["MagneticFieldStrength"],
            )
            self.b0_str = "1.5T"
        else:
            if b0_val == 3:
                self.b0_str = "3.0T"
            else:
                self.b0_str = f"{b0_val}T"

    def order_by(self, images, att):
        """Order images by attribute (e.g. EchoTime, InversionTime)."""
        sorted_images = sorted(images, key=lambda x: x[att].value.real)
        return sorted_images

    def template_fit(self, template_dcm, image_index=0):
        """
        Calculate transformation matrix to fit template to image.

        The template pixel array, template_px, is fitted to one of the
        images in self.images (default=0). The resultant RT matrix is stored as
        self.warp_matrix.

        This matrix can be used to map coordinates from template space to image
        space using transform_coords(...), or to map masks from template space
        to image space using cv2.warpAffine(...).

        To map from image space to template space, the 'inverse' RT matrix can
        be calculated using:
          inverse_transform_matrix = cv.invertAffineTransform(self.warp_matrix)

        Parameters
        ----------
        image_index : int, optional
            Index of image to be used for template matching. The default is 0.

        Returns
        -------
        warp_matrix : np.array
            RT transform matrix (2,3).

        Further details
        ---------------
        Untested for situations where the template matrix is larger than the
        image (lack of data!). Tested for images larger than templates.

        TODO
        ----
        This routine is suboptimal. It may be better to extract the bolt
        hole locations and fit from them, or run an edge-detection algorithm
        as pixel values are highly variable between scanners and manufacturers.

        Need to check if image is real valued, typically signed then shifted so
        background is 2048, or magnitude image. Currently it forces converts
        all images to magnitude images before regression.

        Despite these limitations, this method works well in practice for small
        angle rotations.
        """
        # Store template pixel array, after scaling in 0028,1052 and
        # 0028,1053 applied
        template_px = pixel_rescale(template_dcm)
        target_px = pixel_rescale(self.images[0])

        ## Pad template or target pixels if required
        # Determine difference in shape
        pad_size = np.subtract(template_px.shape, target_px.shape)
        assert pad_size[0] == pad_size[1], "Image matrices must be square."
        if pad_size[0] > 0:  # pad target--UNTESTED
            # add pixels to target if smaller than template
            target_px = np.pad(target_px, pad_width=(0, pad_size[0]))
        elif pad_size[0] < 0:  # pad template
            # add pixels to template if smaller than target
            template_px = np.pad(template_px, pad_width=(0, -pad_size[0]))

        # Always fit on magnitude images for simplicity. May be suboptimal
        self.template8bit = cv.normalize(
            abs(template_px),
            None,
            0,
            255,
            norm_type=cv.NORM_MINMAX,
            dtype=cv.CV_8U,
        )

        self.target8bit = cv.normalize(
            abs(target_px),
            None,
            0,
            255,
            norm_type=cv.NORM_MINMAX,
            dtype=cv.CV_8U,
        )

        # initialise transformation fitting parameters.
        scale_factor = len(target_px) / len(template_px)
        scale_matrix = scale_factor * np.eye(2, 3, dtype=np.float32)

        self.scaled_template8bit = cv.warpAffine(
            self.template8bit,
            scale_matrix,
            (self.template8bit.shape[1], self.template8bit.shape[0]),
        )

        # Apply transformation
        criteria = (
            cv.TERM_CRITERIA_EPS | cv.TERM_CRITERIA_COUNT,
            TEMPLATE_FIT_ITERS,
            TERMINATION_EPS,
        )
        # Find the geometric transform (warp) between two images in terms of the ECC criterion
        self.template_cc, warp_matrix = cv.findTransformECC(
            self.template8bit, self.target8bit, scale_matrix, criteria=criteria
        )

        self.warped_template8bit = cv.warpAffine(
            self.template8bit,
            warp_matrix,
            (self.template8bit.shape[1], self.template8bit.shape[0]),
        )

        return warp_matrix

    def generate_time_series(
        self, coords_row_col, warp_matrix, fit_coords=True
    ):
        """
        Create list of ROITimeSeries objects.

        Parameters
        ----------
        coords_row_col : array_like
            Array of coordinates points of interest (POIs) for each centre of
            each ROI. They should be in [[col0, row0], [col1, row1], ...]
            format.
        time_attribute : string, optional
            DICOM attribute to order images. Typically 'InversionTime' for T1
            relaxometry or 'EchoTime' for T2.
        warp_matrix : np.array
            RT transform matrix (2,3).
        """
        num_coords = np.size(coords_row_col, axis=0)

        # adjustment may not be required for the template DICOM
        if fit_coords:
            adjusted_coords_row_col = transform_coords(
                coords_row_col,
                warp_matrix,
                input_row_col=True,
                output_row_col=True,
            )
        else:  # used in testing
            adjusted_coords_row_col = coords_row_col

        self.ROI_time_series = []
        for i in range(num_coords):
            self.ROI_time_series.append(
                ROITimeSeries(
                    self.images, adjusted_coords_row_col[i], self.time_attr
                )
            )

    def plot_fit(self):
        """
        Visual representation of target fitting.

        Create 2x2 subplot showing 8-bit version of:
            1. Template
            2. Original image
            3. Overlay of (1) and (2)
            4. Overlay of RT transformed template and (2)
        """
        fig = plt.figure()
        plt.subplot(2, 2, 1)
        plt.imshow(self.template8bit, cmap="gray")
        plt.title("Template")
        plt.axis("off")

        plt.subplot(2, 2, 2)
        self.plot_rois(new_fig=False)
        plt.title("Image")

        plt.subplot(2, 2, 3)
        plt.imshow(
            self.scaled_template8bit / 2 + self.target8bit / 2, cmap="gray"
        )
        plt.title("Image / template overlay")
        plt.axis("off")

        plt.subplot(2, 2, 4)
        plt.imshow(
            self.warped_template8bit / 2 + self.target8bit / 2, cmap="gray"
        )
        plt.title("Image / fitted template overlay")
        plt.axis("off")

        plt.tight_layout()

        return fig

    def plot_rois(self, new_fig=True):
        """
        Plot ROIs on image for visual check on template fitting.

        Parameters
        ----------
        new_fig : bool, optional
            Create new figure if True. Otherwise create in current axis (e.g.
            as a subplot). The default is True.

        Returns
        -------
            matplotlib figure handle if a new figure was created, otherwise
            None.

        """
        fig = None
        if new_fig:
            # Create image in a new figure (not a subplot)
            fig = plt.figure()

        plt.imshow(self.target8bit, cmap="gray")
        plt.axis("off")
        if hasattr(self, "ROI_time_series"):
            combined_ROI_map = np.zeros_like(self.ROI_time_series[0].ROI_mask)
            for roi in self.ROI_time_series:
                combined_ROI_map += roi.ROI_mask
            lines = outline_mask(combined_ROI_map)
            for line in lines:
                plt.plot(line[1], line[0], color="r", alpha=1)

        return fig

    @property
    def relax_times(self):
        """List of T1 for each ROI."""
        return [fit[0][0] for fit in self.relax_fit]


class T1ImageStack(ImageStack):
    """
    Calculate T1 relaxometry.
    """

    def __init__(self, image_slices):
        time_attribute = "InversionTime"
        super().__init__(image_slices, time_attribute)

    def generate_t1_function(
        self, ti_interp_vals, tr_interp_vals, mag_image=False
    ):
        """
        Generate T1 signal function and jacobian with interpolated TRs.

        Signal intensity on T1 decay is a function of both TI and TR. Ideally, TR
        should be constant and at least 5*T1. However, scan time can be reduced by
        allowing a shorter TR which increases at long TIs. For example::
            TI |   50 |  100 |  200 |  400 |  600 |  800
            ---+------+------+------+------+------+------
            TR | 1000 | 1000 | 1000 | 1260 | 1860 | 2460

        This function factory returns a function which calculates the signal
        magnitude using the expression::
            S = S0 * (1 - a1 * np.exp(-TI / t1) + np.exp(-TR / t1))
        where ``S0`` is the recovered intensity, ``a1`` is theoretically 2.0 but
        varies due to inhomogeneous B0 field, ``t1`` is the longitudinal
        relaxation time, and the repetition time, ``TR``, is calculated  from
        ``TI`` using piecewise linear interpolation.

        Parameters
        ----------
        ti_interp_vals : array_like
            Array of TI values used as a look-up table to calculate TR
        tr_interp_vals : array_like
            Array of TR values used as a lookup table to calculate TR from the TI
            used in the sequence.
        mag_image : bool, optional
            If True, the generated function returns the magnitude of the signal
            (i.e. negative outputs become positive). The default is False.

        Returns
        -------
        t1_function : function
            S = S0 * (1 - a1 * np.exp(-TI / t1) + np.exp(-TR / t1))

        t1_jacobian : function
            Tuple of partial derivatives for curve fitting.

        eqn_str : string
            String representation of fit function.
        """
        #  Create piecewise linear fit function. k=1 gives linear, s=0 ensures all
        #  points are on line. Using UnivariateSpline (rather than numpy.interp())
        #  enables derivative calculation if required.
        tr = UnivariateSpline(ti_interp_vals, tr_interp_vals, k=1, s=0)
        # tr_der = tr.derivative()

        eqn_str = "s0 * (1 - a1 * np.exp(-TI / t1) + np.exp(-TR / t1))"
        if mag_image:
            eqn_str = f"abs({eqn_str})"

        def _t1_function_signed(ti, t1, s0, a1):
            pv = s0 * (1 - a1 * np.exp(-ti / t1) + np.exp(-tr(ti) / t1))
            return pv

        def t1_function(ti, t1, s0, a1):
            pv = _t1_function_signed(ti, t1, s0, a1)
            if mag_image:
                return abs(pv)
            else:
                return pv

        def t1_jacobian(ti, t1, s0, a1):
            t1_der = (
                s0
                / (t1**2)
                * (-ti * a1 * np.exp(-ti / t1) + tr(ti) * np.exp(-tr(ti) / t1))
            )
            s0_der = 1 - a1 * np.exp(-ti / t1) + np.exp(-tr(ti) / t1)
            a1_der = -s0 * np.exp(-ti / t1)
            jacobian = np.array([t1_der, s0_der, a1_der])

            if mag_image:
                pv = _t1_function_signed(ti, t1, s0, a1)
                jacobian = (jacobian * (pv >= 0)) - (jacobian * (pv < 0))

            return jacobian.T

        return t1_function, t1_jacobian, eqn_str

    def generate_fit_function(self):
        """ "Create T1 fit function for magnitude/signed image and variable TI."""
        #  check if image is signed or magnitude
        if np.all(pixel_rescale(self.images[0]) >= 0):
            mag_image = True
        else:
            mag_image = False
        (
            self.fit_function,
            self.fit_jacobian,
            self.fit_eqn_str,
        ) = self.generate_t1_function(
            self.ROI_time_series[0].times,
            self.ROI_time_series[0].trs,
            mag_image=mag_image,
        )

    def est_t1_s0(self, ti, tr, t1, pv):
        """
        Return initial guess of s0 to seed T1 curve fitting.

        Parameters
        ----------
        ti : array_like
            TI values.
        tr : array_like
            TR values.
        t1 : array_like
            Estimated T1 (typically from manufacturer's documentation).
        pv : array_like
            Mean pixel value (signal) in ROI.

        Returns
        -------
        array_like
            Initial s0 guess for calculating T1 relaxation time.

        """
        return -pv / (1 - 2 * np.exp(-ti / t1) + np.exp(-tr / t1))

    def initialise_fit_parameters(self, t1_estimates):
        """
        Estimate fit parameters (t1, s0, a1) for T1 curve fitting.

        T1 estimates are provided.

        s0 is estimated using abs(est_t1_s0(ti, tr, t1_est, mean_pv))
            For each ROI, s0 is calculated using from both the smallest and
            largest TI, and the value with the largest mean_pv used. This
            guards against the case where division by a mean_pv close to zero
            causes a large rounding error.

        A1 is estimated as 2.0, the theoretical value assuming homogeneous B0

        Parameters
        ----------
        t1_estimates : array_like
            T1 values to seed estimation. These should be the manufacturer
            provided T1 values where known.

        Returns
        -------
        s0_est

        """
        self.t1_est = t1_estimates
        rois = self.ROI_time_series
        rois_first_mean = np.array([roi.means[0] for roi in rois])
        rois_last_mean = np.array([roi.means[-1] for roi in rois])
        s0_est_last = abs(
            self.est_t1_s0(
                rois[0].times[-1], rois[0].trs[-1], self.t1_est, rois_last_mean
            )
        )
        s0_est_first = abs(
            self.est_t1_s0(
                rois[0].times[0], rois[0].trs[0], self.t1_est, rois_first_mean
            )
        )
        s0_est = np.where(
            rois_first_mean > rois_last_mean, s0_est_first, s0_est_last
        )
        self.a1_est = np.full_like(s0_est, 2.0)

        return s0_est

    def find_relax_times(self, t1_estimates, s0_est):
        """
        Calculate T1 values. Access as ``image_stack.relax_fits``

        Parameters
        ----------
        t1_estimates : array_like
            T1 values to seed estimation. These should be the manufacturer
            provided T1 values where known.

        Returns
        -------
        None.

        """
        rois = self.ROI_time_series
        self.relax_fit = [
            scipy.optimize.curve_fit(
                self.fit_function,
                rois[i].times,
                rois[i].means,
                p0=[t1_estimates[i], s0_est[i], self.a1_est[i]],
                jac=self.fit_jacobian,
                method="lm",
            )
            for i in range(len(rois))
        ]


class T2ImageStack(ImageStack):
    """
    Calculate T2 relaxometry.
    """

    def __init__(self, image_slices):
        time_attribute = "EchoTime"
        super().__init__(image_slices, time_attribute)

        self.fit_eqn_str = "T2 with Rician noise (Raya et al 2010)"

    def generate_fit_function(self):
        """Null method in base class, may be overwritten in subclass."""

    def fit_function(self, te, t2, s0, c):
        """
        Calculated pixel value with Rician noise model.

        Calculates pixel value from [1]_::
            .. math::
                S=sqrt{frac{pi alpha^2}{2}} exp(- alpha) left( (1+ 2 alpha)
                text{I_0}(alpha) + 2 alpha text{I_1}(alpha) right)

                alpha() = left(frac{S_0}{2 sigma} exp{left(-frac{text{TE}}{text{T}_2}right)} right)^2

                text{I}_n() = n^text{th} text{order modified Bessel function of the first kind}

        Parameters
        ----------
        te : array_like
            Echo times.
        t2 : float
            T2 decay constant.
        S0 : float
            Initial pixel magnitude.
        C : float
            Noise parameter for Rician model (equivalent to st dev).

        Returns
        -------
        pv : array_like
            Theoretical pixel values (signal) at each TE.

        References
        ----------
        .. [1] Raya, J.G., Dietrich, O., Horng, A., Weber, J., Reiser, M.F. and
        Glaser, C., 2010. T2 measurement in articular cartilage: impact of the
        fitting method on accuracy and precision at low SNR. Magnetic Resonance in
        Medicine: An Official Journal of the International Society for Magnetic
        Resonance in Medicine, 63(1), pp.181-193.
        """

        alpha = (s0 / (2 * c) * np.exp(-te / t2)) ** 2
        # NB need to use `i0e` and `ive` below to avoid numeric inaccuracy from
        # multiplying by huge exponentials then dividing by the same exponential
        pv = np.sqrt(np.pi / 2 * c**2) * (
            (1 + 2 * alpha) * i0e(alpha) + 2 * alpha * ive(1, alpha)
        )

        return pv

    def est_t2_s0(self, te, t2, pv, c=0.0):
        """
        Initial guess for s0 to seed curve fitting::
            .. math::
                S_0=\\frac{pv-c}{exp(-TE/T_2)}


        Parameters
        ----------
        te : array_like
            Echo time(s).
        t2 : array_like
            T2 decay constant.
        pv : array_like
            Mean pixel value (signal) in ROI with ``te`` echo time.
        c : array_like
            Constant offset, theoretically ``full_like(te, 0.0)``.

        Returns
        -------
        array_like
            Initial s0 estimate.

        """
        return (pv - c) / np.exp(-te / t2)

    def initialise_fit_parameters(self, t2_estimates):
        """
        Estimate fit parameters (t2, s0, c) for T2 curve fitting.

        T2 estimates are provided.

        s0 is estimated using est_t2_s0(te, t2_est, mean_pv, c).

        C is estimated as 5.0.

        Parameters
        ----------
        t2_estimates : array_like
            T2 values to seed estimation. These should be the manufacturer
            provided T2 values where known.

        Returns
        -------
        s0_est.

        """
        self.t2_est = t2_estimates
        rois = self.ROI_time_series
        rois_second_mean = np.array([roi.means[1] for roi in rois])
        self.c_est = np.full_like(self.t2_est, SEED_RICIAN_NOISE)
        # estimate s0 from second image--first image is too low.
        s0_est = self.est_t2_s0(
            rois[0].times[1], self.t2_est, rois_second_mean, self.c_est
        )

        return s0_est

    def find_relax_times(self, t2_estimates, s0_est):
        """
        Calculate T2 values. Access as ``image_stack.relax_times``

        Uses the 'skip first echo' fit method [1]_ with a Rician noise model
        [2]_. Ideally the Rician noise parameter should be determined from the
        images rather than fitted. However, this is not possible as the noise
        profile varies across the image due to spatial coil sensitivities and
        whether the image is normalised or unfiltered. Fitting the noise
        parameter makes this easier. It has an upper limit of MAX_RICIAN_NOISE,
        currently set to 20.0.

        Parameters
        ----------
        t2_estimates : array_like
            T2 values to seed estimation. These should be the manufacturer
            provided T2 values where known.

        Returns
        -------
        None.

        References
        ----------
        .. [1] McPhee, K. C., & Wilman, A. H. (2018). Limitations of skipping
        echoes for exponential T2 fitting. Journal of Magnetic Resonance
        Imaging, 48(5), 1432-1440. https://doi.org/10.1002/jmri.26052

        .. [2] Raya, J.G., Dietrich, O., Horng, A., Weber, J., Reiser, M.F. and
        Glaser, C., 2010. T2 measurement in articular cartilage: impact of the
        fitting method on accuracy and precision at low SNR. Magnetic Resonance
        in Medicine: An Official Journal of the International Society for
         Magnetic Resonance in Medicine, 63(1), pp.181-193.
        """

        rois = self.ROI_time_series
        #  Omit the first image data from the curve fit. This is achieved by
        #  slicing rois[i].times[1:] and rois[i].means[1:].
        #  Skipping odd echoes can be implemented with
        #  rois[i].times[1::2] and .means[1::2]

        bounds = ([0, 0, 1], [np.inf, np.inf, MAX_RICIAN_NOISE])

        self.relax_fit = [
            scipy.optimize.curve_fit(
                self.fit_function,
                rois[i].times[1:],
                rois[i].means[1:],
                p0=[t2_estimates[i], s0_est[i], self.c_est[i]],
                jac=None,
                bounds=bounds,
                method="trf",
            )
            for i in range(len(rois))
        ]



"""Low-Contrast Object Detectability.

As per section 7 of the Large and Medium Phantom Test Guidance for the
ACR MRI Accreditation Program:

https://www.acraccreditation.org/-/media/ACRAccreditation/Documents/MRI/ACR-Large-Med-Phantom-Guidance-102022.pdf

```
the low contrast detectability test assesses the extent to which objects
of low contrast are discernible in the images.
```

These are performed on slices 8 through 11 by counting
the number of visible spokes.

The implementation follows that of:

A statistical approach to automated analysis of the
low-contrast object detectability test for the large ACR MRI phantom

DOI = {10.1002/acm2.70173}
journal = {Journal of Applied Clinical Medical Physics},
author = {Golestani, Ali M. and Gee, Julia M.},
year = {2025},
month = jul

An implementation by the authors can be found on GitHub:
    https://github.com/aligoles/ACR-Low-Contrast-Object-Detectability

With the original paper:
    https://doi.org/10.1002/acm2.70173
    https://aapm.onlinelibrary.wiley.com/doi/pdfdirect/10.1002/acm2.70173

ACR Low Contrast Object Detectability:
    https://mriquestions.com/uploads/3/4/5/7/34572113/largephantomguidance.pdf

Notes from the paper:

- Images with acquired with:
        - 3.0T Siemens MAGNETOM Vida.
        - 1.5T Philips scanner integrated into an Elekta Unity MR-Linac System.
- 40 Datasets analyzed (20 for each scanner).


Paper Implementation Overview:

- Normalise image intensity for each slice (independently) to within [0, 1].
- Background removal process performed using histogram thresholding.
- Contrast disk is identified.
        - Detect center of phantom, crop and then find a large circle.
- 90 Angular radials profile in a specific angle are generated.
- Known phantom geometry and rotation used to calculate position of first spoke.
        - Circles at 12.5, 25.0 and 38.0mm from CoG.
- 2nd order polynomial fitted to the model and added to general linear model
    (GLM) regressors.
- 3 GLM regressors created for each 1D profile.
- Test passes if every GLM regressors exceed the significance level.
        - Significance level for each slice is set to 0.0125.
- Significance within each slice is adjusted using the Benjamini-Hochberg
    false discovery rate.

Hazen Implementation Overview:

- Normalise image intensity for each slice (independently) to within [0, 1].
- Phantom background removal via binary thresholding and morphological erosion.
- LCOD disc center detected using Hough Circle transform on cropped gradient images.
        - Average center coordinates across slices 8-11 for sub-pixel stability.
- Geometric template generated per slice with 9-degree rotational offsets between slices.
        - 10 radial spokes with objects at 12.5, 25.0 and 38.0mm from center.
        - Optional ensemble template matching testing +- 3 pixel spatial perturbations.
- 128-sample radial intensity profiles extracted along each spoke trajectory.
- 3rd order polynomial detrending using robust fitting truncated at 90% profile length.
- Gaussian smoothing (sigma=5) applied to detrended profiles.
- GLM fitted to each profile with object mask and constant regressors.
        - Parameters estimated using statsmodels GLM.
        - Ensemble selection based on minimum summed p-values with positive coefficients.
- Benjamini-Hochberg FDR correction applied across all 30 tests per slice (10 spokes x 3 objects).
- Detection requires FDR-corrected p-value < alpha (default 0.05) and positive GLM parameter.
- Pass thresholds applied based on magnetic field strength.
        - 1.5T: minimum 7 spokes detected across all slices.
        - 3.0T: minimum 37 spokes detected across all slices.

Key Differences from Paper:

- Uses Hough Circle detection rather than histogram thresholding for disc identification.
- Implements ensemble template optimization to handle sub-pixel registration errors.
- Employs 3rd order polynomial detrending rather than 2nd order.
- Uses standard Benjamini-Hochberg FDR rather than fixed alpha=0.0125 per slice.
- Field-strength specific pass criteria (7 vs 37 spokes) based on ACR guidance.
- Simplified background masking via morphological operations rather than complex histogram analysis.
- Gaussian smoothing (sigma=5) applied pre-GLM fitting rather than raw profiles.

Implemented for Hazen by Alex Drysdale: alexander.drysdale@wales.nhs.uk
"""

# Typing
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import pydicom

# Python imports
import contextlib
import copy
import logging
from pathlib import Path

# Module imports
import cv2
import matplotlib as mpl
import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import numpy as np
import scipy as sp
import skimage.transform
import statsmodels
import statsmodels.api as sm
from matplotlib.gridspec import GridSpec
from matplotlib.patches import Circle

from hazenlib.ACRObject import ACRObject
from hazenlib.HazenTask import HazenTask
from hazenlib.types import (
    FailedStatsModel,
    LCODTemplate,
    Measurement,
    P_HazenTask,
    Result,
    SpokeReportData,
    StatsParameters,
)

logger = logging.getLogger(__name__)


class ACRLowContrastObjectDetectability(HazenTask):
    """Low Contrast Object Detectability (LCOD) class for the ACR phantom.

    Attributes:
        SLICE_ANGLE_OFFSET : Angular offset between each subsequent slice
                in degrees. Defaults to 9.
        START_ANGLE : Starting angle for slice 0 in degrees.
                Defaults to 0.
        LCOD_DISC_SIZE : LCOD disc radius in mm.

    """

    SLICE_ANGLE_OFFSET: float = 9
    START_ANGLE: float = 0
    LCOD_DISC_SIZE: float = 43  # mm

    _DETREND_POLYNOMIAL_ORDER: int = 3
    _STD_TOL: float = 0.01

    _SMOOTH_SIGMA: float = 3

    _RADIAL_PROFILE_LENGTH: int = 128
    # Paper states a value of 0.0125
    # https://aapm.onlinelibrary.wiley.com/doi/pdfdirect/10.1002/acm2.70173
    # However this is too high since the this implementation includes
    # a much more robust center finding and template matching strategy.
    _ALPHA: float = 1e-6

    # Control how templates are sample from around the center.
    # Note that 0 (i.e. the center) is always sampled.
    _ENSEMBLE_MAX_INC: float = (
        2  # Max increment from the center in units of dx
    )
    _ENSEMBLE_STEP: float = 0.1  # Steps between increments in units of dx.

    OBJECT_RIBBON_COLORS = ("#1E88E5", "#FFC107", "#004D40")

    def __init__(
        self,
        alpha: float | None = None,
        **kwargs: P_HazenTask.kwargs,
    ) -> None:
        """Initialise the LCOD object."""
        if kwargs.pop("verbose", None) is not None:
            logger.warning(
                "verbose is not a supported argument for %s",
                type(self).__name__,
            )
        super().__init__(**kwargs)

        self.alpha = self._ALPHA if alpha is None else alpha

        # Start at last slice (highest contrast) and work backwards
        self.slice_range = slice(10, 6, -1)

        # Initialise ACR object
        self.ACR_obj = ACRObject(self.dcm_list)
        self.rotation = self.ACR_obj.determine_rotation(
            self.ACR_obj.slice_stack[0].pixel_array,
        )
        self.lcod_center = None

        # Pass threshold is at least N spokes total for both the T1 and T2
        # acquisitions where:
        # @ 1.5T, N =  7
        # @ 3.0T, N = 37
        match float(
            self.ACR_obj.slice_stack[0]["MagneticFieldStrength"].value,
        ):
            case 3.0:
                self.pass_threshold = 37
            case 1.5:
                self.pass_threshold = 7
            case _:
                logger.error(
                    "No LCOD pass threshold specified for %s T systems"
                    " assuming a pass threshold of at least 7 spokes for"
                    " each sequence",
                    self.ACR_obj.slice_stack[0]["MagneticFieldStrength"].value,
                )
                self.pass_threshold = 7

        # Only used in reporting
        self.slice_report_data = {}

    def run(self) -> Result:
        """Run the LCOD analysis."""
        results = self.init_result_dict(desc=self.ACR_obj.acquisition_type())

        results.files = [
            self.img_desc(f)
            for f in self.ACR_obj.slice_stack[self.slice_range]
        ]

        total_spokes = 0
        # TODO(abdrysdale) : Use wait_on_parallel_results to collect results.
        for i, dcm in enumerate(self.ACR_obj.slice_stack[self.slice_range]):
            slice_no = 1 + self.slice_range.step * i + self.slice_range.start
            result = self.count_spokes(
                dcm,
                slice_no=slice_no,
                alpha=self.alpha,
            )
            try:
                num_spokes = min(i for i, r in enumerate(result) if not r)
            except ValueError:
                num_spokes = len(result)

            # Add individual spoke measurements for debugging
            # and further analysis
            # If this results in log flooding then it might be best to remove
            for j, r in enumerate(result):
                spoke_no = j + 1
                results.add_measurement(
                    Measurement(
                        name="LowContrastObjectDetectability",
                        type="measured",
                        subtype=f"slice {slice_no} spoke {spoke_no}",
                        value=r,
                        visibility="intermediate",
                    ),
                )
            total_spokes += num_spokes
            results.add_measurement(
                Measurement(
                    name="LowContrastObjectDetectability",
                    type="measured",
                    subtype=f"slice {slice_no}",
                    value=num_spokes,
                ),
            )

        results.add_measurement(
            Measurement(
                name="LowContrastObjectDetectability",
                type="measured",
                subtype="total",
                value=total_spokes,
            ),
        )

        results.add_measurement(
            Measurement(
                name="LowContrastObjectDetectability",
                type="measured",
                subtype="pass/fail",
                value=total_spokes >= self.pass_threshold,
            ),
        )

        if self.report:
            results.add_report_image(self.report_files)

        return results

    def _get_template_ensemble(
        self,
        template: LCODTemplate,
    ) -> list[LCODTemplate]:
        cx, cy, theta = template.cx, template.cy, template.theta
        max_inc = self._ENSEMBLE_MAX_INC
        num = np.ceil((max_inc * 2) / self._ENSEMBLE_STEP).astype(int) + 1
        increments = list(
            np.linspace(
                -max_inc * self.ACR_obj.dx,
                +max_inc * self.ACR_obj.dx,
                num,
            ),
        )

        if not np.any(np.isclose(increments, 0.0)):
            increments.append(0.0)

        return sorted(
            [
                LCODTemplate(
                    cx + dx,
                    cy + dy,
                    theta,
                )
                for dx in increments
                for dy in increments
            ],
            key=lambda t: t.cx**2 + t.cy**2,
        )

    def _get_params_and_p_vals(
        self,
        template: LCODTemplate,
        dcm: pydicom.Dataset,
        *,
        use_ensemble_template: bool = False,
    ) -> StatsParameters:
        """Get a list of parameters and associated p-values."""
        sp = StatsParameters()
        if use_ensemble_template:
            templates = self._get_template_ensemble(template)
        else:
            templates = [template]

        for idx, _ in enumerate(template.spokes):
            min_pvals = [1, 1, 1]  # Initial p values
            for _template in templates:
                spoke = _template.spokes[idx]
                profile, object_mask = spoke.profile(
                    dcm,
                    size=self._RADIAL_PROFILE_LENGTH,
                    return_object_mask=True,
                )

                p_vals, params = self._analyze_profile(
                    profile,
                    object_mask,
                )

                if (
                    np.sum(p_vals) < np.sum(min_pvals)  # noqa: F821
                    and all(params > 0)
                ):
                    min_pvals = p_vals
                    min_params = params

                # As a rule of thumb, if all of the p-values
                # are a factor of 10 less than the alpha value
                # then the spoke has passed even with FDR correction
                # this may need to be changed to a higher factor
                # such as 100 if evidence is provided on the contrary.
                # This avoids excessive searching of templates once the
                # it has already been determined that the spoke has passed
                # but without performing fdr correction on every template.
                if all(p < self.alpha / 10 for p in min_pvals):
                    break
            sp.p_vals.append(min_pvals)
            sp.params.append(min_params)

        return sp

    def _fdrcorrection(
        self,
        sp: StatsParameters,
        alpha: float,
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        rejected, p_vals_fdr = statsmodels.stats.multitest.fdrcorrection(
            sp.p_vals_all,
            alpha=alpha,
            method="indep",
            is_sorted=False,
        )
        rejected = rejected.reshape(-1, len(sp.p_vals[-1]))
        params_fdr = np.array(sp.params_all).reshape(-1, len(sp.params[-1]))
        p_vals_fdr = p_vals_fdr.reshape(-1, len(sp.p_vals[-1]))

        return (rejected, params_fdr, p_vals_fdr)

    def count_spokes(
        self,
        raw: pydicom.Dataset,
        slice_no: int = -1,
        alpha: float = 0.05,
    ) -> np.ndarray:
        """Count spokes with optional report data capture."""
        dcm = self._preprocess(raw)
        template = self.get_current_slice_template(slice_no)
        spokes = template.spokes

        # Get analysis data
        report_data = [] if self.report else None

        for spoke_id, spoke in enumerate(spokes):
            profile, (x_coords, y_coords), object_mask = spoke.profile(
                dcm,
                size=self._RADIAL_PROFILE_LENGTH,
                return_coords=True,
                return_object_mask=True,
            )

            # Analyze profile
            if self.report:
                p_vals, params, detrended, trend = self._analyze_profile(
                    profile,
                    object_mask,
                    return_intermediate=True,
                )
                # Store data for reporting
                report_data.append(
                    SpokeReportData(
                        spoke_id=spoke_id,
                        profile=profile,
                        detrended=detrended,
                        trend=trend,
                        object_mask=object_mask,
                        x_coords=x_coords,
                        y_coords=y_coords,
                        p_vals=p_vals,
                        params=params,
                        detected=[],  # Will be filled after FDR correction
                        objects=spoke.objects,
                    ),
                )

        sp = self._get_params_and_p_vals(
            template,
            dcm,
            use_ensemble_template=True,
        )

        # FDR correction
        rejected, params_fdr, p_vals_fdr = self._fdrcorrection(sp, alpha=alpha)

        # Update detection status
        for spoke_number, spoke in enumerate(spokes):
            for i, obj in enumerate(spoke):
                obj.detected = (
                    rejected[spoke_number, i]
                    and params_fdr[spoke_number, i] > 0
                )
            spoke.passed = all(obj.detected for obj in spoke)

            # Update report data with detection status
            if self.report:
                report_data[spoke_number].detected = [
                    obj.detected for obj in spoke
                ]
                report_data[spoke_number].p_vals = p_vals_fdr[spoke_number, :]
                report_data[spoke_number].params = params_fdr[spoke_number, :]

        # Store report data if reporting enabled
        if self.report:
            self.slice_report_data[slice_no] = report_data
            self.generate_slice_report(dcm, slice_no, report_data)

        return [s.passed for s in spokes]

    def find_center(self, crop_ratio: float = 0.55) -> tuple[float]:
        """Find the center of the LCOD phantom."""
        if self.lcod_center is not None:
            return self.lcod_center

        # Get ACR Phantom Center
        (main_cx, main_cy), main_radius = self.ACR_obj.find_phantom_center(
            self.ACR_obj.slice_stack[0].pixel_array,
            self.ACR_obj.dx,
            self.ACR_obj.dy,
        )

        # Get cropped image of LCOD disk
        r = main_radius * crop_ratio

        lcod_r_init = self.LCOD_DISC_SIZE  # mm
        lcod_cy = main_cy + 5 / self.ACR_obj.dy

        offset_y = max(0, int(lcod_cy - r))
        offset_x = max(0, int(main_cx - r))

        # Gets LCOD center and radius from each slice.
        lcod_props = []
        for dcm in self.ACR_obj.slice_stack[8:]:
            cropped_image = dcm.pixel_array[
                offset_y : int(lcod_cy + r + 1),
                offset_x : int(main_cx + r + 1),
            ]
            cropped_image = (
                (cropped_image - cropped_image.min())
                * 255.0
                / (cropped_image.max() - cropped_image.min())
            ).astype(np.uint8)

            # Pre-processing for circle detection
            img_blur = cv2.GaussianBlur(cropped_image, (5, 5), 0)
            img_grad = img_blur.max() - img_blur

            try:
                detected_circles = cv2.HoughCircles(
                    img_grad,
                    method=cv2.HOUGH_GRADIENT,
                    dp=2,
                    minDist=cropped_image.shape[0] // 2,
                    minRadius=int((lcod_r_init - 2) / self.ACR_obj.dy),
                    maxRadius=int((lcod_r_init + 2) / self.ACR_obj.dy),
                ).flatten()
            except AttributeError:
                logger.warning("Failed to find LCOD center, using defaults.")
                detected_circles = (main_cx, lcod_cy, lcod_r_init)
                lcod_center = detected_circles[:2]

            else:
                lcod_center = tuple(
                    (dc + offset) * dv
                    for dc, offset, dv in zip(
                        detected_circles[:2],
                        (offset_x, offset_y),
                        (self.ACR_obj.dx, self.ACR_obj.dy),
                        strict=True,
                    )
                )

            lcod_props.append((lcod_center, detected_circles[2]))

        # Obtains the mean lcod center and radius.
        lcod_center = (
            np.mean([prop[0][0] for prop in lcod_props]),
            np.mean([prop[0][1] for prop in lcod_props]),
        )
        lcod_r = np.mean([prop[1] for prop in lcod_props])
        self.lcod_center = lcod_center

        if self.report:
            fig, axes = plt.subplots(2, 2, constrained_layout=True)

            # Initial Estimate
            axes[0, 0].imshow(dcm.pixel_array, cmap="gray")
            axes[0, 0].scatter(
                main_cx / self.ACR_obj.dx,
                lcod_cy / self.ACR_obj.dy,
                marker="x",
                color="red",
            )
            circle = Circle(
                (
                    main_cx / self.ACR_obj.dx,
                    lcod_cy / self.ACR_obj.dy,
                ),
                lcod_r_init / self.ACR_obj.dx,
                fill=False,
                edgecolor="red",
                linewidth=0.5,
            )
            axes[0, 0].add_patch(circle)
            axes[0, 0].set_title("Initial Estimate")

            # Cropped
            axes[0, 1].imshow(cropped_image)
            axes[0, 1].scatter(
                main_cx / self.ACR_obj.dx - offset_x,
                lcod_cy / self.ACR_obj.dy - offset_y,
                marker="x",
                color="red",
            )
            circle = Circle(
                (
                    main_cx / self.ACR_obj.dx - offset_x,
                    lcod_cy / self.ACR_obj.dy - offset_y,
                ),
                lcod_r_init / self.ACR_obj.dx,
                fill=False,
                edgecolor="red",
                linewidth=1,
            )
            axes[0, 1].add_patch(circle)
            axes[0, 1].set_title("Cropped Image")

            # Initial guess
            cx, cy = detected_circles[:2]
            circle = Circle(
                (cx / self.ACR_obj.dx, cy / self.ACR_obj.dy),
                lcod_r / self.ACR_obj.dx,
                fill=False,
                edgecolor="blue",
                linewidth=2,
            )
            axes[1, 0].imshow(cropped_image, cmap="gray")
            axes[1, 0].scatter(
                detected_circles[0],
                detected_circles[1],
                marker="x",
                color="blue",
            )
            axes[1, 0].add_patch(circle)
            axes[1, 0].set_title("Detected Circle")

            # Final guess
            cx = self.lcod_center[0] / self.ACR_obj.dx - offset_x
            cy = self.lcod_center[1] / self.ACR_obj.dy - offset_y
            circle = Circle(
                (cx, cy),
                lcod_r / self.ACR_obj.dx,
                fill=False,
                edgecolor="blue",
                linewidth=2,
            )
            axes[1, 1].imshow(cropped_image, cmap="gray")
            axes[1, 1].scatter(
                cx,
                cy,
                marker="x",
                color="blue",
            )
            axes[1, 1].add_patch(circle)
            axes[1, 1].set_title(
                "Detected Circle - Final",
            )

            fig.suptitle("LCOD Center Detection")

            data_path = Path(self.dcm_list[0].filename).parent.name
            img_path = (
                Path(self.report_path)
                / f"{data_path}_center_{self.img_desc(dcm)}.png"
            )
            fig.savefig(img_path)
            plt.close()
            self.report_files.append(img_path)

        return lcod_center

    def _current_slice_rotation(self, current_slice: int) -> float:
        rotation_offset = self.START_ANGLE + self.SLICE_ANGLE_OFFSET * abs(
            current_slice - 8,
        )
        return self.rotation + rotation_offset

    def get_current_slice_template(
        self,
        current_slice: int,
    ) -> LCODTemplate:
        """Find the position of the spokes within the LCOD disk."""
        # Rotation offset
        theta = self._current_slice_rotation(current_slice)

        if self.lcod_center is None:
            cx_0, cy_0 = self.find_center()  # updates lcod_center
        else:
            cx_0, cy_0 = self.lcod_center

        template = LCODTemplate(cx_0, cy_0, theta)

        logger.info(
            "Template generated for slice %i:"
            "\nCenter:\t(%f, %f)"
            "\nRotation:\t%f (initial: %f + offset: %f)",
            current_slice,
            template.cx,
            template.cy,
            template.theta,
            self.rotation,
            template.theta - self.rotation,
        )

        return template

    def _analyze_profile(
        self,
        profile: np.ndarray,
        object_mask: np.ndarray,
        *,
        mask_padding: int = 5,
        return_intermediate: bool = False,
    ) -> tuple:
        """Analyze radial profile with optional intermediate returns."""
        # Apply binary dilation (mask padding)
        footprint_n = mask_padding * 2 + 1
        object_mask = skimage.morphology.binary_dilation(
            object_mask,
            footprint=np.array(
                [[0] * footprint_n, [1] * footprint_n, [0] * footprint_n],
            ).T,
        )

        # De-trend with robust polynomial fitting
        detrended, trend = self._detrend_profile(profile, return_trend=True)
        smoothed = self._smooth_profile(detrended, sigma=self._SMOOTH_SIGMA)

        # Prepare GLM
        data = np.column_stack((object_mask, np.ones_like(profile)))

        # Fit GLM
        try:
            model = sm.GLM(smoothed, data).fit()
        except ValueError:
            logger.exception("Fit could not be obtained - failing detection")
            model = FailedStatsModel()

        if return_intermediate:
            return model.pvalues[:3], model.params[:3], smoothed, trend
        return model.pvalues[:3], model.params[:3]

    # Add the report generation method
    def generate_slice_report(
        self,
        dcm: pydicom.Dataset,
        slice_no: int,
        report_data: list[SpokeReportData],
    ) -> None:
        """Generate comprehensive report for a single slice."""
        fig = plt.figure(figsize=(20, 15))
        gs = GridSpec(3, 4, figure=fig, hspace=0.2, wspace=0.2)

        # Main image (top-left, spans 1 row and 1 column)
        ax_main = fig.add_subplot(gs[0, 0])
        self._plot_main_image(ax_main, dcm, report_data, slice_no)

        # Profile plots for each spoke
        n_spokes = len(report_data)
        for i, spoke_data in enumerate(report_data):
            if i > n_spokes:  # Limit to available grid space (11 slots max)
                break

            # Calculate grid position (skip main image at [0,0])
            row = (i + 1) // 4
            col = (i + 1) % 4

            # Create broken axis within this grid cell
            self._create_broken_axis_plot(fig, gs[row, col], spoke_data)

        # Summary table (bottom row, last column)
        ax_table = fig.add_subplot(gs[2, 3])
        self._plot_summary_table(ax_table, report_data)

        legend_elements = [
            mpatches.Patch(
                facecolor=self.OBJECT_RIBBON_COLORS[0],
                alpha=0.4,
                label="Inner Object",
            ),
            mpatches.Patch(
                facecolor=self.OBJECT_RIBBON_COLORS[1],
                alpha=0.4,
                label="Middle Object",
            ),
            mpatches.Patch(
                facecolor=self.OBJECT_RIBBON_COLORS[2],
                alpha=0.4,
                label="Outer Object",
            ),
        ]
        ax_main.legend(
            handles=legend_elements,
            loc="lower left",
            fontsize=6,
            framealpha=0.9,
        )

        # Save figure
        data_path = Path(self.dcm_list[0].filename).parent.name
        img_path = (
            Path(self.report_path)
            / f"{data_path}_lcod_slice_{str(slice_no).zfill(2)}.png"
        )
        fig.savefig(img_path, dpi=150, bbox_inches="tight")
        plt.close(fig)
        self.report_files.append(img_path)

    def _plot_main_image(
        self,
        ax: plt.Axes,
        dcm: pydicom.Dataset,
        report_data: list[SpokeReportData],
        slice_no: int,
    ) -> None:
        """Plot cropped image focused on LCOD object region."""
        px_x, px_y = self.ACR_obj.dx, self.ACR_obj.dy

        # Calculate bounds of all objects across all spokes
        all_x_coords = [
            obj.x for spoke_data in report_data for obj in spoke_data.objects
        ]
        all_y_coords = [
            obj.y for spoke_data in report_data for obj in spoke_data.objects
        ]

        # Determine crop region with 10% margin
        min_x, max_x = min(all_x_coords), max(all_x_coords)
        min_y, max_y = min(all_y_coords), max(all_y_coords)
        margin_x = (max_x - min_x) * 0.1
        margin_y = (max_y - min_y) * 0.1

        # Convert to pixel coordinates
        x_min = int(max(0, (min_x - margin_x) / px_x))
        x_max = int(min(dcm.pixel_array.shape[1], (max_x + margin_x) / px_x))
        y_min = int(max(0, (min_y - margin_y) / px_y))
        y_max = int(min(dcm.pixel_array.shape[0], (max_y + margin_y) / px_y))

        # Crop image
        cropped_img = dcm.pixel_array[y_min:y_max, x_min:x_max]

        vmin, vmax = self._window(dcm)
        # Display cropped region
        ax.imshow(
            cropped_img,
            cmap="gray",
            vmin=vmin,
            vmax=vmax,
        )

        # Adjust drawing coordinates for crop offset
        offset_x = x_min * px_x
        offset_y = y_min * px_y

        # Draw overlays
        for spoke_data in report_data:
            if (
                spoke_data.x_coords is not None
                and spoke_data.y_coords is not None
            ):
                ax.plot(
                    (spoke_data.x_coords - x_min),
                    (spoke_data.y_coords - y_min),
                    "y-",
                    linewidth=1.5,
                    alpha=0.8,
                )

            for obj, detected in zip(
                spoke_data.objects,
                spoke_data.detected,
                strict=True,
            ):
                color = "green" if detected else "red"
                circle = Circle(
                    ((obj.x - offset_x) / px_x, (obj.y - offset_y) / px_y),
                    (obj.diameter / 2) / px_x,
                    fill=False,
                    edgecolor=color,
                    linewidth=0.5,
                )
                ax.add_patch(circle)

        ax.set_title(f"Slice {slice_no}", fontsize=10, fontweight="bold")
        ax.axis("off")

        # Add legend
        detected_patch = mpatches.Patch(color="green", label="Detected")
        not_detected_patch = mpatches.Patch(color="red", label="Not Detected")
        ax.legend(
            handles=[detected_patch, not_detected_patch],
            loc="upper right",
            fontsize=6,
        )

    def _create_broken_axis_plot(
        self,
        fig: plt.Figure,
        gs_cell: mpl.gridspec.SubplotSpec,
        spoke_data: SpokeReportData,
    ) -> None:
        """Create broken axis plot with prominent object ribbons."""
        cell_bbox = gs_cell.get_position(fig)
        top_height = cell_bbox.height * 0.5
        bottom_height = cell_bbox.height * 0.5

        ax_top = fig.add_axes(
            [
                cell_bbox.x0,
                cell_bbox.y0 + bottom_height,
                cell_bbox.width,
                top_height,
            ],
        )
        ax_bottom = fig.add_axes(
            [cell_bbox.x0, cell_bbox.y0, cell_bbox.width, bottom_height],
            sharex=ax_top,
        )

        x = np.arange(len(spoke_data.profile))

        # Plot profiles
        ax_top.plot(
            x,
            spoke_data.profile,
            "k-",
            label="Original",
            linewidth=1,
            zorder=3,
        )
        ax_top.plot(
            x,
            spoke_data.trend,
            "g--",
            label="Trend",
            linewidth=0.8,
            zorder=3,
        )
        ax_bottom.plot(x, spoke_data.detrended, "k-", linewidth=1, zorder=3)

        # Set y-limits
        orig_min, orig_max = (
            np.min(spoke_data.profile),
            np.max(spoke_data.profile),
        )
        det_min, det_max = (
            np.min(spoke_data.detrended),
            np.max(spoke_data.detrended),
        )
        orig_range = orig_max - orig_min
        det_range = det_max - det_min
        ax_top.set_ylim(
            orig_min - 0.05 * orig_range,
            orig_max + 0.05 * orig_range,
        )
        ax_bottom.set_ylim(
            det_min - 0.05 * det_range,
            det_max + 0.05 * det_range,
        )

        # Style axes
        ax_top.spines["bottom"].set_visible(False)
        ax_top.xaxis.tick_top()
        ax_top.tick_params(labeltop=False, labelsize=6)
        ax_bottom.spines["top"].set_visible(False)
        ax_bottom.xaxis.tick_bottom()
        ax_bottom.tick_params(labelbottom=True, labelsize=6)

        # Diagonal break markers
        d = 0.5
        kwargs = {
            "marker": [(-1, -d), (1, d)],
            "markersize": 8,
            "linestyle": "none",
            "color": "k",
            "mec": "k",
            "mew": 1,
            "clip_on": False,
        }
        ax_top.plot([0, 1], [0, 0], transform=ax_top.transAxes, **kwargs)
        ax_bottom.plot([0, 1], [1, 1], transform=ax_bottom.transAxes, **kwargs)

        # **Enhanced vertical ribbons for objects**
        n_objects = spoke_data.object_mask.shape[1]
        for obj_idx in range(n_objects):
            obj_mask = spoke_data.object_mask[
                :,
                obj_idx,
            ]  # Get mask for this object
            color = self.OBJECT_RIBBON_COLORS[obj_idx]

            obj_indices = np.where(obj_mask)[0]
            if len(obj_indices) > 0:
                # Expand region for prominence (20% padding)
                pad = max(1, len(obj_indices) // 5)
                start = max(0, obj_indices[0] - pad)
                end = min(len(x), obj_indices[-1] + pad)

                # Draw prominent vertical ribbons (behind profile lines)
                ax_top.axvspan(start, end, alpha=0.45, color=color, zorder=0)
                ax_bottom.axvspan(
                    start,
                    end,
                    alpha=0.45,
                    color=color,
                    zorder=0,
                )

                # Add subtle edge
                ax_top.axvline(
                    start,
                    color=color,
                    alpha=0.7,
                    linewidth=0.5,
                    zorder=1,
                )
                ax_top.axvline(
                    end,
                    color=color,
                    alpha=0.7,
                    linewidth=0.5,
                    zorder=1,
                )
                ax_bottom.axvline(
                    start,
                    color=color,
                    alpha=0.7,
                    linewidth=0.5,
                    zorder=1,
                )
                ax_bottom.axvline(
                    end,
                    color=color,
                    alpha=0.7,
                    linewidth=0.5,
                    zorder=1,
                )

                # Annotate p-value at object center
                mid_x = int(np.mean(obj_indices))
                mid_y = spoke_data.detrended[mid_x] + spoke_data.trend[mid_x]
                ax_bottom.annotate(
                    f"p={spoke_data.p_vals[obj_idx]:.3f}",
                    xy=(mid_x, mid_y),
                    xytext=(3, 3),
                    textcoords="offset points",
                    fontsize=5,
                    ha="left",
                    bbox={
                        "boxstyle": "round,pad=0.2",
                        "facecolor": "yellow",
                        "alpha": 0.85,
                        "edgecolor": "none",
                    },
                    zorder=4,
                )

        # Labels
        ax_bottom.set_xlabel("Profile Position", fontsize=7)
        ax_top.set_ylabel("Original", fontsize=7)
        ax_bottom.set_ylabel("Detrended & Smoothed", fontsize=7)

        # Title
        status = "PASS" if all(spoke_data.detected) else "FAIL"
        n_detected = sum(spoke_data.detected)
        ax_top.set_title(
            f"Spoke {spoke_data.spoke_id + 1}:"
            f" {n_detected}/{len(spoke_data.detected)} {status}",
            fontsize=8,
            fontweight="bold",
        )

        ax_top.grid(visible=True, alpha=0.3)
        ax_bottom.grid(visible=True, alpha=0.3)

    def _plot_summary_table(
        self,
        ax: plt.Axes,
        report_data: list[SpokeReportData],
    ) -> None:
        """Plot condensed summary table (one row per spoke)."""
        ax.axis("off")

        # One row per spoke, objects shown as columns
        table_data = []
        headers = [
            "Spoke",
            "O1 p-val",
            "O1 param",
            "O2 p-val",
            "O2 param",
            "O3 p-val",
            "O3 param",
        ]

        for spoke_data in report_data:
            row = [f"{spoke_data.spoke_id + 1}"]
            for p_val, param in zip(
                spoke_data.p_vals,
                spoke_data.params,
                strict=True,
            ):
                # Compact formatting
                p_str = (
                    f"{p_val:.2e}"
                    if p_val < 1e-3  # noqa: PLR2004
                    else f"{p_val:.4f}"
                )
                param_str = f"{param:.3f}"
                row.extend([p_str, param_str])

            table_data.append(row)

        # Create table
        cw = 0.16
        table = ax.table(
            cellText=table_data,
            colLabels=headers,
            cellLoc="center",
            loc="center",
            colWidths=[0.10, cw, cw, cw, cw, cw, cw],
        )

        # Style
        table.auto_set_font_size(False)  # noqa: FBT003
        table.set_fontsize(6)  # Smaller font for compactness
        table.scale(1, 1.8)

        # Header styling
        for i, _ in enumerate(headers):
            table[(0, i)].set_facecolor("#4472C4")
            table[(0, i)].set_text_props(weight="bold", color="white")

        # Color cells: light green for pass, light red for fail
        f_color = "#FAADAD"
        p_color = "#C6E0B4"
        for i, spoke_data in enumerate(report_data):
            for j in range(3):  # For each of 3 objects
                is_fail = spoke_data.p_vals[j] > self._ALPHA
                color = f_color if is_fail else p_color
                table[(i + 1, j * 2 + 1)].set_facecolor(color)  # p-value cell

                is_fail = spoke_data.params[j] <= 0
                color = f_color if is_fail else p_color
                table[(i + 1, j * 2 + 2)].set_facecolor(
                    color,
                )  # parameter cell

    def _window(
        self,
        dcm: pydicom.FileDataset,
        idx: int | None = None,
        squeeze: float = 2,
    ) -> tuple[float]:
        """Return vmin, vmax values based on simple window method."""
        (cx, cy) = self.lcod_center
        r = (self.LCOD_DISC_SIZE - squeeze) * self.ACR_obj.dx
        y_grid, x_grid = np.meshgrid(
            np.arange(0, dcm.pixel_array.shape[0]),
            np.arange(0, dcm.pixel_array.shape[1]),
        )

        mask = (y_grid - cy) ** 2 + (x_grid - cx) ** 2 <= r**2

        try:
            vdata = (
                dcm.pixel_array * mask[None, :, :]
                if idx is None
                else dcm.pixel_array[idx, :, :] * mask
            )
        except IndexError:
            vdata = dcm.pixel_array * mask

        return (np.min(vdata[vdata != 0]), np.max(vdata))

    @staticmethod
    def _preprocess(
        dcm: pydicom.FileDataset,
        threshold: tuple[float] = (0.05, 0.65),
        threshold_step: float = 0.001,
        bounds: tuple[float] = (0.1, 0.2),
    ) -> pydicom.FileDataset:
        """Preprocess the DICOM."""
        processed = copy.deepcopy(dcm)
        data = processed.pixel_array

        # Normalise
        fdata = data / np.max(data)

        # Threshold
        threshold_min, threshold_max = threshold
        lower, upper = bounds
        structure = np.ones((3, 3), dtype=int)
        for thr in np.arange(threshold_min, threshold_max, threshold_step):
            _, thresh = cv2.threshold(fdata, thr, 1, 0)
            labelled, _ = sp.ndimage.label(
                thresh,
                structure,
            )
            thresh_inner = labelled == np.max(labelled)
            if lower < np.sum(thresh_inner != 0) / np.sum(fdata != 0) < upper:
                break

        # Erode the circular mask
        data *= skimage.morphology.erosion(
            thresh_inner,
        )

        processed.set_pixel_data(
            data,
            dcm[(0x0028, 0x0004)].value,  # Photometric Interpretation
            dcm[(0x0028, 0x0101)].value,  # Bits Stored
        )
        return processed

    def _smooth_profile(
        self,
        profile: np.ndarray,
        sigma: int = 5,
    ) -> np.ndarray:
        return sp.ndimage.gaussian_filter(
            profile,
            sigma=sigma,
            mode="constant",
            cval=np.mean(profile),
        )

    def _detrend_profile(
        self,
        profile_potentially_unsigned: np.ndarray,
        *,
        return_trend: bool = False,
    ) -> np.ndarray:
        """Remove polynomial trend from profile using robust fitting."""
        # Ignores end of profile which sometimes includes outer disc.
        profile = profile_potentially_unsigned.astype(np.int16)
        try:
            stop_idx = np.where(
                profile <= profile.max() * 0.75,
            )[0][0]

            if stop_idx <= len(profile) * 0.9:
                stop_idx = int(len(profile) * 0.9)

        except IndexError:
            stop_idx = len(profile)

        trunc_profile = profile[:stop_idx]

        if np.std(trunc_profile) > self._STD_TOL:
            x = np.linspace(0, 1, len(trunc_profile))
            coeffs = np.polyfit(
                x,
                trunc_profile,
                self._DETREND_POLYNOMIAL_ORDER,
            )
            trend = np.zeros_like(profile)
            trend[:stop_idx] = np.polyval(coeffs, x)
            trend[stop_idx:] = profile[stop_idx:]
        else:
            trend = np.zeros_like(profile)

        detrend = profile - trend

        if return_trend:
            return (detrend, trend)
        return detrend

    def _plot_profile_debug(
        self,
        detrended: np.ndarray,
        profile: np.ndarray,
        peak_indices: np.ndarray,
        intersection_points: list[tuple[float, float]],
        spoke_id: int,
        slice_no: int,
        object_mask: np.ndarray,
        x_coords: np.ndarray,
        y_coords: np.ndarray,
    ) -> None:
        """Create debug plot for radial profile analysis.

        Visualizes the profile data, detected peaks, half-maximum intersection
        points, and object mask regions to help verify the intersection point
        detection algorithm.

        Args:
            detrended: Detrended but not smoothed profile.
            profile: Smoothed and detrended 1D intensity profile
            peak_indices: Array of indices where peaks were detected
            intersection_points: List of (x, y) coordinate tuples for
                half-max points in image space
            spoke_id: Spoke number (0-9)
            slice_no: Slice number (8-11)
            object_mask: Binary mask array indicating object locations (size x 3)
            x_coords: X coordinates in image space for each profile point
            y_coords: Y coordinates in image space for each profile point

        """
        if not self.report:
            return

        # Convert intersection points from image coordinates back to profile indices
        # by finding the closest point in the profile coordinates
        intersection_indices = []
        with contextlib.suppress(TypeError):
            for xi, yi in intersection_points:
                distances = np.sqrt(
                    (x_coords - xi) ** 2 + (y_coords - yi) ** 2,
                )
                closest_idx = np.argmin(distances)
                intersection_indices.append(closest_idx)

        # Create figure
        fig, ax = plt.subplots(figsize=(10, 4))

        # Profile positions (0 to length-1)
        profile_positions = np.arange(len(profile))

        # Plot the smoothed profile
        ax.plot(
            profile_positions,
            profile,
            color="#1976D2",
            linewidth=1.5,
            label="Smoothed + Detrended profile",
        )
        ax.plot(
            profile_positions,
            detrended,
            color="#19D2A4",
            linewidth=1.5,
            linestyle="--",
            label="Detrended profile",
        )

        # Mark detected peaks
        if len(peak_indices) > 0:
            ax.scatter(
                peak_indices,
                profile[peak_indices],
                color="#D32F2F",
                marker="^",
                s=80,
                zorder=5,
                label="Peaks",
            )

        # Mark half-maximum intersection points
        if len(intersection_indices) > 0:
            # Ensure indices are within bounds
            valid_indices = [
                idx for idx in intersection_indices if 0 <= idx < len(profile)
            ]
            if valid_indices:
                ax.scatter(
                    valid_indices,
                    profile[np.array(valid_indices).astype(int)],
                    color="#388E3C",
                    marker="o",
                    s=60,
                    zorder=5,
                    label="Half-max points",
                )

        # Add vertical bands for object mask regions
        if object_mask is not None and object_mask.shape[1] > 0:
            for obj_idx in range(object_mask.shape[1]):
                obj_mask_col = object_mask[:, obj_idx]
                # Find contiguous regions where mask is True
                mask_start = None
                for i, is_masked in enumerate(obj_mask_col):
                    if is_masked and mask_start is None:
                        mask_start = i
                    elif not is_masked and mask_start is not None:
                        # End of masked region
                        color = self.OBJECT_RIBBON_COLORS[
                            obj_idx % len(self.OBJECT_RIBBON_COLORS)
                        ]
                        ax.axvspan(
                            mask_start,
                            i,
                            alpha=0.15,
                            color=color,
                            zorder=1,
                        )
                        mask_start = None
                # Handle case where mask extends to end
                if mask_start is not None:
                    color = self.OBJECT_RIBBON_COLORS[
                        obj_idx % len(self.OBJECT_RIBBON_COLORS)
                    ]
                    ax.axvspan(
                        mask_start,
                        len(obj_mask_col),
                        alpha=0.15,
                        color=color,
                        zorder=1,
                    )

        # Set axis limits with padding
        profile_range = np.ptp(detrended)
        if profile_range > 0:
            ax.set_ylim(
                detrended.min() - 0.1 * profile_range,
                detrended.max() + 0.1 * profile_range,
            )

        # Labels and title
        ax.set_xlabel("Profile Position", fontsize=8)
        ax.set_ylabel("Intensity", fontsize=8)
        ax.set_title(
            f"LCOD Profile Debug - Slice {slice_no}, Spoke {spoke_id}",
            fontsize=9,
        )
        ax.legend(fontsize=7, loc="best")
        ax.tick_params(labelsize=7)
        ax.grid(True, alpha=0.3, linestyle="--", linewidth=0.5)  # noqa: FBT003

        plt.tight_layout()

        # Save figure
        data_path = Path(self.dcm_list[0].filename).parent.name
        img_path = (
            Path(self.report_path)
            / f"{data_path}_profile_debug_slice{slice_no}_spoke{spoke_id}.png"
        )
        fig.savefig(img_path, dpi=100, bbox_inches="tight")
        plt.close(fig)
        self.report_files.append(img_path)



"""
ACR Geometric Accuracy

https://www.acraccreditation.org/-/media/acraccreditation/documents/mri/largephantomguidance.pdf

Calculates geometric accuracy for slices 1 and 5 of the ACR phantom.

This script calculates the horizontal and vertical lengths of the ACR phantom in Slice 1 in accordance with the ACR
Guidance.
This script calculates the horizontal, vertical and diagonal lengths of the ACR phantom in Slice 5 in accordance with
the ACR Guidance.
The average distance measurement error, maximum distance measurement error and coefficient of variation of all distance
measurements is reported as recommended by IPEM Report 112, "Quality Control and Artefacts in Magnetic Resonance
Imaging".

This is done by first producing a binary mask for each respective slice. Line profiles are drawn with aid of rotation
matrices around the centre of the test object to determine each respective length. The results are also visualised.

Created by Yassine Azma
yassine.azma@rmh.nhs.uk

18/11/2022
"""

import os
import sys
import traceback

import numpy as np
import skimage.measure
import skimage.morphology
import skimage.transform
from hazenlib.ACRObject import ACRObject
from hazenlib.HazenTask import HazenTask
from hazenlib.logger import logger
from hazenlib.types import Measurement, Result


class ACRGeometricAccuracy(HazenTask):
    """Geometric accuracy measurement class for DICOM images of the ACR phantom."""

    def __init__(self, **kwargs):
        if kwargs.pop("verbose", None) is not None:
            logger.warning(
                "verbose is not a supported argument for %s",
                type(self).__name__,
            )
        super().__init__(**kwargs)
        self.ACR_obj = ACRObject(self.dcm_list)

    def run(self) -> Result:
        """Main function for performing geometric accuracy measurement using the first and fifth slices from the ACR phantom image set.

        Returns:
            dict: results are returned in a standardised dictionary structure specifying the task name, input DICOM Series Description + SeriesNumber + InstanceNumber, task measurement key-value pairs, optionally path to the generated images for visualisation.
        """

        # Initialise results dictionary
        results = self.init_result_dict(desc=self.ACR_obj.acquisition_type())
        results.files = [
            self.img_desc(self.ACR_obj.slice_stack[0]),
            self.img_desc(self.ACR_obj.slice_stack[4]),
        ]

        try:
            lengths_1 = self.get_geometric_accuracy(0)
            for len_1, t in zip(lengths_1, ("Horizontal", "Vertical")):
                results.add_measurement(
                    Measurement(
                        name="GeometricAccuracy",
                        type="measured",
                        subtype=f"Slice 1 {t} distance",
                        description=self.img_desc(self.ACR_obj.slice_stack[0]),
                        value=round(len_1, 2),
                        unit="mm",
                    ),
                )

        except Exception as e:
            logger.exception(
                "Could not calculate the geometric accuracy for"
                " %s because of : %s",
                self.img_desc(self.ACR_obj.slice_stack[0]),
                e,
            )
            traceback.print_exc(file=sys.stdout)

        try:
            lengths_5 = self.get_geometric_accuracy(4)
            for len_5, t in zip(
                lengths_5,
                (
                    "Horizontal",
                    "Vertical",
                    "Diagonal SW",
                    "Diagonal SE",
                ),
            ):
                results.add_measurement(
                    Measurement(
                        name="GeometricAccuracy",
                        type="measured",
                        subtype=f"Slice 5 {t} distance",
                        description=self.img_desc(self.ACR_obj.slice_stack[4]),
                        value=round(len_5, 2),
                        unit="mm",
                    ),
                )
        except Exception as e:
            logger.exception(
                "Could not calculate the geometric accuracy for"
                " %s because of : %s",
                self.img_desc(self.ACR_obj.slice_stack[4]),
                e,
            )
            traceback.print_exc(file=sys.stdout)

        L = lengths_1 + lengths_5

        mean_err, max_err, cov_l = self.get_distortion_metrics(L)

        results.add_measurement(
            Measurement(
                name="GeometricAccuracy",
                type="measured",
                subtype="distortion",
                description="Mean relative error",
                value=round(mean_err, 2),
                visibility="intermediate",
            ),
        )
        results.add_measurement(
            Measurement(
                name="GeometricAccuracy",
                type="measured",
                subtype="distortion",
                description="Max absolute error",
                value=round(max_err, 2),
                visibility="intermediate",
            ),
        )
        results.add_measurement(
            Measurement(
                name="GeometricAccuracy",
                type="fitted",
                subtype="distortion",
                description="Coefficient of variation",
                value=round(cov_l, 2),
                unit="%",
                visibility="intermediate",
            ),
        )

        # only return reports if requested
        if self.report:
            results.add_report_image(self.report_files)

        return results

    def get_geometric_accuracy(self, slice_index):
        """Measure geometric accuracy for input slice. \n
        Creates a mask over the phantom from the pixel array of the DICOM image.
        Uses the centre and shape of the mask to determine horizontal and vertical lengths,
        and also diagonal lengths in slice 5.

        Args:
            slice_index (int): the index of the slice position, for example slice 5 is at index 4.

        Returns:
            tuple of float: horizontal and vertical distances.
        """
        img_dcm = self.ACR_obj.slice_stack[slice_index]
        img = img_dcm.pixel_array
        cxy, _ = self.ACR_obj.find_phantom_center(
            img, self.ACR_obj.dx, self.ACR_obj.dy
        )
        mask = self.ACR_obj.get_mask_image(img, cxy)

        length_dict = self.ACR_obj.measure_orthogonal_lengths(mask, cxy)
        if slice_index == 4:
            sw_dict, se_dict = self.diagonal_lengths(mask, cxy, 4)

        if self.report:
            import matplotlib.pyplot as plt

            fig, axes = plt.subplots(3, 1)
            fig.set_size_inches(8, 24)
            fig.tight_layout(pad=4)

            if slice_index == 0:
                axes[0].imshow(img)
                axes[0].scatter(cxy[0], cxy[1], c="red")
                axes[0].set_title("Centroid Location")

                axes[1].imshow(mask)
                axes[1].set_title("Thresholding Result")

                axes[2].imshow(img)
                axes[2].arrow(
                    length_dict["Horizontal Extent"][0],
                    cxy[1],
                    length_dict["Horizontal Extent"][-1]
                    - length_dict["Horizontal Extent"][0],
                    1,
                    color="blue",
                    length_includes_head=True,
                    head_width=5,
                )
                axes[2].arrow(
                    cxy[0],
                    length_dict["Vertical Extent"][0],
                    1,
                    length_dict["Vertical Extent"][-1]
                    - length_dict["Vertical Extent"][0],
                    color="orange",
                    length_includes_head=True,
                    head_width=5,
                )
                axes[2].legend(
                    [
                        str(np.round(length_dict["Horizontal Distance"], 2))
                        + "mm",
                        str(np.round(length_dict["Vertical Distance"], 2))
                        + "mm",
                    ]
                )
                axes[2].axis("off")
                axes[2].set_title("Geometric Accuracy for Slice 1")

                img_path = os.path.realpath(
                    os.path.join(
                        self.report_path, f"{self.img_desc(img_dcm)}.png"
                    )
                )
                fig.savefig(img_path)
                self.report_files.append(img_path)

            if slice_index == 4:
                axes[0].imshow(img)
                axes[0].scatter(cxy[0], cxy[1], c="red")
                axes[0].axis("off")
                axes[0].set_title("Centroid Location")

                axes[1].imshow(mask)
                axes[1].axis("off")
                axes[1].set_title("Thresholding Result")

                axes[2].imshow(img)
                axes[2].arrow(
                    length_dict["Horizontal Extent"][0],
                    cxy[1],
                    length_dict["Horizontal Extent"][-1]
                    - length_dict["Horizontal Extent"][0],
                    1,
                    color="blue",
                    length_includes_head=True,
                    head_width=5,
                )
                axes[2].arrow(
                    cxy[0],
                    length_dict["Vertical Extent"][0],
                    1,
                    length_dict["Vertical Extent"][-1]
                    - length_dict["Vertical Extent"][0],
                    color="orange",
                    length_includes_head=True,
                    head_width=5,
                )
                axes[2].arrow(
                    se_dict["Start"][0],
                    se_dict["Start"][1],
                    se_dict["Extent"][0],
                    se_dict["Extent"][1],
                    color="purple",
                    length_includes_head=True,
                    head_width=5,
                )
                axes[2].arrow(
                    sw_dict["Start"][0],
                    sw_dict["Start"][1],
                    sw_dict["Extent"][0],
                    sw_dict["Extent"][1],
                    color="yellow",
                    length_includes_head=True,
                    head_width=5,
                )

                axes[2].legend(
                    [
                        str(np.round(length_dict["Horizontal Distance"], 2))
                        + "mm",
                        str(np.round(length_dict["Vertical Distance"], 2))
                        + "mm",
                        str(np.round(sw_dict["Distance"], 2)) + "mm",
                        str(np.round(se_dict["Distance"], 2)) + "mm",
                    ]
                )
                axes[2].axis("off")
                axes[2].set_title("Geometric Accuracy for Slice 5")

                img_path = os.path.realpath(
                    os.path.join(
                        self.report_path, f"{self.img_desc(img_dcm)}.png"
                    )
                )
                fig.savefig(img_path)
                self.report_files.append(img_path)

        if slice_index == 4:
            return (
                length_dict["Horizontal Distance"],
                length_dict["Vertical Distance"],
                sw_dict["Distance"],
                se_dict["Distance"],
            )
        else:
            return length_dict["Horizontal Distance"], length_dict[
                "Vertical Distance"
            ]

    def diagonal_lengths(self, img, cxy, slice_index):
        """Measure diagonal lengths. \n
        Rotates the pixel array by 45° and measures the horizontal and vertical distances.

        Args:
            img (np.ndarray): pixel array of the slice (dcm.pixel_array).
            cxy (tuple): x,y coordinates of the circle centre.
            slice_index (int): index of the slice number.

        Returns:
            tuple of dictionaries: for both the south-east (SE) diagonal length and the south-west (SW) diagonal length: \n
            "start" and "end" indicate the start and end x and y positions of the lengths; "Extent" is the distance (in
            pixels) of the lengths; "Distance" is "Extent" with factors applied to convert from pixels to mm.
        """
        # Calculate geometric mean of the x and y pixel spacing components,
        # due to the possibility of pixels being rectangular,
        # ie. the length and width of pixels can differ.
        eff_res = np.sqrt(
            np.mean(np.square((self.ACR_obj.dx, self.ACR_obj.dy)))
        )
        img_rotate = skimage.transform.rotate(img, 45, center=(cxy[0], cxy[1]))

        length_dict = self.ACR_obj.measure_orthogonal_lengths(img_rotate, cxy)
        extent_h = length_dict["Horizontal Extent"]

        origin = (cxy[0], cxy[1])
        start = (extent_h[0], cxy[1])
        end = (extent_h[-1], cxy[1])
        se_x_start, se_y_start = ACRObject.rotate_point(origin, start, 45)
        se_x_end, se_y_end = ACRObject.rotate_point(origin, end, 45)

        dist_se = (
            np.sqrt(
                np.sum(
                    np.square([se_x_end - se_x_start, se_y_end - se_y_start])
                )
            )
            * eff_res
        )
        se_dict = {
            "Start": (se_x_start, se_y_start),
            "End": (se_x_end, se_y_end),
            "Extent": (se_x_end - se_x_start, se_y_end - se_y_start),
            "Distance": dist_se,
        }

        extent_v = length_dict["Vertical Extent"]

        start = (cxy[0], extent_v[0])
        end = (cxy[0], extent_v[-1])
        sw_x_start, sw_y_start = ACRObject.rotate_point(origin, start, 45)
        sw_x_end, sw_y_end = ACRObject.rotate_point(origin, end, 45)

        dist_sw = (
            np.sqrt(
                np.sum(
                    np.square([sw_x_end - sw_x_start, sw_y_end - sw_y_start])
                )
            )
            * eff_res
        )
        sw_dict = {
            "Start": (sw_x_start, sw_y_start),
            "End": (sw_x_end, sw_y_end),
            "Extent": (sw_x_end - sw_x_start, sw_y_end - sw_y_start),
            "Distance": dist_sw,
        }

        return sw_dict, se_dict

    @staticmethod
    def get_distortion_metrics(L):
        """Calculates the mean error, the maximum error and the coefficient of
        variation between the horizontal and vertical distances
        measured on slices 1 and 5.

        Args:
            L (tuple): horizontal and vertical distances from slices 1 and 5.

        Returns:
            tuple of floats: mean_err, max_err, cov_l
        """
        err = [x - 190 for x in L]
        mean_err = np.mean(err)

        max_err = np.max(np.absolute(err))
        cov_l = 100 * np.std(L) / np.mean(L)

        return mean_err, max_err, cov_l



"""
Spatial Resolution

Contributors:
Haris Shuaib, haris.shuaib@gstt.nhs.uk
Neil Heraghty, neil.heraghty@nhs.net, 16/05/2018

.. todo::
    Replace shape finding functions with hazenlib.utils equivalents

"""

import copy
import os
import sys
import traceback

import cv2 as cv
import numpy as np
from hazenlib.HazenTask import HazenTask
from hazenlib.logger import logger
from hazenlib.types import Measurement, Result
from hazenlib.utils import get_pe_direction, get_pixel_size, rescale_to_byte
from numpy.fft import fftfreq


class SpatialResolution(HazenTask):
    """Spatial resolution measurement class for DICOM images of the MagNet phantom

    Inherits from HazenTask class
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.single_dcm = self.dcm_list[0]

    def run(self) -> Result:
        """Main function for performing spatial resolution measurement

        Returns:
            dict: results are returned in a standardised dictionary structurespecifying the task name, input DICOM Series Description + SeriesNumber + InstanceNumber, task measurement key-value pairs, optionally path to the generated images for visualisation
        """
        results = self.init_result_dict()
        results.files = self.img_desc(self.single_dcm)
        try:
            pe_result, fe_result = self.calculate_mtf(self.single_dcm)
            results.add_measurement(
                Measurement(
                    name="SpatialResolution",
                    type="measured",
                    subtype="phase encoding direction",
                    unit="mm",
                    value=round(pe_result, 2),
                ),
            )
            results.add_measurement(
                Measurement(
                    name="SpatialResolution",
                    subtype="frequency encoding direction",
                    unit="mm",
                    value=round(fe_result, 2),
                ),
            )
        except Exception as e:
            logger.exception(
                "Could not calculate the spatial resolution for %s"
                " because of : %s",
                self.img_desc(self.single_dcm),
                e,
            )
            traceback.print_exc(file=sys.stdout)

        # only return reports if requested
        if self.report:
            results.add_report_image(self.report_files)

        return results

    def get_circles(self, image):
        """Locate Hough Circles in a DICOM pixel array

        Args:
            image (array): DICOM pixel array rescaled to byte

        Returns:
            np.array: pixel array of the located circle
        """
        # TODO: use shape functions from utils
        v = np.median(image)
        upper = int(min(255, (1.0 + 5) * v))
        i = 40

        while True:
            circles = cv.HoughCircles(
                image,
                cv.HOUGH_GRADIENT,
                1.2,
                256,
                param1=upper,
                param2=i,
                minRadius=80,
                maxRadius=200,
            )
            # min and max radius need to accomodate at least 256 and 512 matrix sizes
            i -= 1
            if circles is None:
                pass
            else:
                circles = np.uint16(np.around(circles))
                break

        # img = cv.circle(image, (circles[0][0][0], circles[0][0][1]), circles[0][0][2], (255, 0, 0))
        # plt.imshow(img)
        # plt.show()
        return circles

    def thresh_image(self, img, bound=150):
        """Create a threshold image

        Args:
            img (np.array): pixel array
            bound (int, optional): _description_. Defaults to 150.

        Returns:
            np.array: thresholded pixel array
        """
        blurred = cv.GaussianBlur(img, (5, 5), 0)
        thresh = cv.threshold(blurred, bound, 255, cv.THRESH_TOZERO_INV)[1]
        return thresh

    def find_square(self, img):
        # TODO: use shape functions from utils
        cnts = cv.findContours(
            img.copy(), cv.RETR_LIST, cv.CHAIN_APPROX_SIMPLE
        )[0]

        for c in cnts:
            perimeter = cv.arcLength(c, True)
            approx = cv.approxPolyDP(c, 0.1 * perimeter, True)
            if len(approx) == 4:
                # compute the bounding box of the contour and use the
                # bounding box to compute the aspect ratio
                rect = cv.minAreaRect(approx)

                # OpenCV 4.5 adjustment
                # - cv.minAreaRect() output tuple order changed since v3.4
                # - swap rect[1] order & rotate rect[2] by -90
                #   convert tuple>list>tuple to do this
                rectAsList = list(rect)
                rectAsList[1] = (rectAsList[1][1], rectAsList[1][0])
                rectAsList[2] = rectAsList[2] - 90
                rect = tuple(rectAsList)

                box = cv.boxPoints(rect)
                box = np.intp(box)
                w, h = rect[1]
                ar = w / float(h)

                # make sure that the width of the square is reasonable size taking into account 256 and 512 matrix
                if not 20 < w < 100:
                    continue

                # a square will have an aspect ratio that is approximately
                # equal to one, otherwise, the shape is a rectangle
                if 0.92 < ar < 1.08:
                    break

        # points should start at top-right and go anti-clockwise
        top_corners = sorted(box, key=lambda x: x[1])[:2]
        top_corners = sorted(top_corners, key=lambda x: x[0], reverse=True)

        bottom_corners = sorted(box, key=lambda x: x[1])[2:]
        bottom_corners = sorted(bottom_corners, key=lambda x: x[0])
        return top_corners + bottom_corners, box

    def get_roi(self, pixels, centre, size=20):
        """Get coordinates of the region of interest

        Args:
            pixels (np.array): _description_
            centre (tuple): x,y (int) coordinates
            size (int, optional): diameter of the region of interest. Defaults to 20.

        Returns:
            np.array: subset of the pixel array
        """
        y, x = centre
        arr = pixels[
            x - size // 2 : x + size // 2, y - size // 2 : y + size // 2
        ]
        return arr

    def get_void_roi(self, pixels, circle, size=20):
        """Create an 'empty' region of interest - same size filles with 0

        Args:
            pixels (np.array): _description_
            centre (tuple): x,y (int) coordinates
            size (int, optional): diameter of the region of interest. Defaults to 20.

        Returns:
            np.array: subset of the pixel array
        """
        centre_x = circle[0][0][0]
        centre_y = circle[0][0][1]
        return self.get_roi(
            pixels=pixels, centre=(centre_x, centre_y), size=size
        )

    def get_edge_roi(self, pixels, edge_centre, size=20):
        return self.get_roi(
            pixels, centre=(edge_centre["x"], edge_centre["y"]), size=size
        )
        return self.get_roi(
            pixels, centre=(edge_centre["x"], edge_centre["y"]), size=size
        )

    def edge_is_vertical(self, edge_roi, mean) -> bool:
        """Determine whether edge is vertical
            control_parameter_01=0  ;a control parameter that will be equal to 1 if the edge is vertical and 0 if it is horizontal

        for column=0, event.MTF_roi_size-2 do begin
        if MTF_Data(column, 0 ) EQ mean_value then control_parameter_01=1
        if (MTF_Data(column, 0) LT mean_value) AND (MTF_Data(column+1, 0) GT mean_value) then control_parameter_01=1
        if (MTF_Data(column, 0) GT mean_value) AND (MTF_Data(column+1, 0) LT mean_value) then control_parameter_01=1
        end
        Args:
            edge_roi (np.array): pixel array in ROI
            mean (np.array): array of mean pixel values

        Returns:
            bool: True or false whether edge is vertical
        """
        for col in range(edge_roi.shape[0] - 1):
            if edge_roi[col, 0] == mean:
                return True
            if edge_roi[col, 0] < mean < edge_roi[col + 1, 0]:
                return True
            if edge_roi[col, 0] > mean > edge_roi[col + 1, 0]:
                return True

        return False

    def get_bisecting_normal(self, vector, centre, length_factor=0.25):
        # calculate coordinates of bisecting normal
        nrx_1 = centre["x"] - int(length_factor * vector["y"])
        nry_1 = centre["y"] + int(length_factor * vector["x"])
        nrx_2 = centre["x"] + int(length_factor * vector["y"])
        nry_2 = centre["y"] - int(length_factor * vector["x"])
        return nrx_1, nry_1, nrx_2, nry_2

    def get_top_edge_vector_and_centre(self, square):
        # Calculate dx and dy
        top_edge_profile_vector = {
            "x": (square[0][0] + square[1][0]) // 2,
            "y": (square[0][1] + square[1][1]) // 2,
        }

        # Calculate centre (x,y) of edge
        top_edge_profile_roi_centre = {
            "x": (square[0][0] + square[1][0]) // 2,
            "y": (square[0][1] + square[1][1]) // 2,
        }

        return top_edge_profile_vector, top_edge_profile_roi_centre

    def get_right_edge_vector_and_centre(self, square):
        # Calculate dx and dy
        right_edge_profile_vector = {
            "x": square[3][0] - square[0][0],
            "y": square[3][1] - square[0][1],
        }  # nonsense

        # Calculate centre (x,y) of edge
        right_edge_profile_roi_centre = {
            "x": (square[3][0] + square[0][0]) // 2,
            "y": (square[3][1] + square[0][1]) // 2,
        }
        return right_edge_profile_vector, right_edge_profile_roi_centre

    def get_signal_roi(self, pixels, edge, edge_centre, circle_r, size=20):
        """Get pixel array from the image within ROI

        Args:
            pixels (np.array): _description_
            edge (_type_): _description_
            edge_centre (_type_): _description_
            circle_r (float/int): circle radius
            size (int, optional): diameter of the region of interest. Defaults to 20.

        Returns:
            np.array: subset of the pixel array
        """

        if edge == "right":
            x = edge_centre["x"] + circle_r // 2
            y = edge_centre["y"]
        elif edge == "top":
            x = edge_centre["x"]
            y = edge_centre["y"] - circle_r // 2

        return self.get_roi(pixels=pixels, centre=(x, y), size=size)

    def get_edge(self, edge_arr, mean_value, spacing):
        # TODO: simplify this function
        if self.edge_is_vertical(edge_arr, mean_value):
            edge_arr = np.rot90(edge_arr)

        x_edge = [0] * 20
        y_edge = [0] * 20

        for row in range(20):
            for col in range(19):
                control_parameter_02 = 0

                if edge_arr[row, col] == mean_value:
                    control_parameter_02 = 1
                if (edge_arr[row, col] < mean_value) and (
                    edge_arr[row, col + 1] > mean_value
                ):
                    control_parameter_02 = 1
                if (edge_arr[row, col] > mean_value) and (
                    edge_arr[row, col + 1] < mean_value
                ):
                    control_parameter_02 = 1

                if control_parameter_02 == 1:
                    x_edge[row] = row * spacing[0]
                    y_edge[row] = col * spacing[1]

        return x_edge, y_edge, edge_arr

    def get_edge_angle_and_intercept(self, x_edge, y_edge):
        """Get edge (slope) angle and intercept value

        Args:
            x_edge (np.ndarray): _description_
            y_edge (np.ndarray): _description_

        Returns:
            tuple: angle and intercept values
        """
        # ;%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        # ;Apply least squares method for the edge
        # ;%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

        mean_x = np.mean(x_edge)
        mean_y = np.mean(y_edge)

        slope_up = np.sum((x_edge - mean_x) * (y_edge - mean_y))
        slope_down = np.sum((x_edge - mean_x) * (x_edge - mean_x))
        slope = slope_up / slope_down
        angle = np.arctan(slope)
        intercept = mean_y - slope * mean_x
        return angle, intercept

    def get_edge_profile_coords(self, angle, intercept, spacing):
        """translate and rotate the data's coordinates according to the slope and intercept

        Args:
            angle (ndarray or scalar): angle of slope
            intercept (ndarray or scalar): intercept of slope
            spacing (tuple/list): spacing value in x and y directions

        Returns:
            tuple: of np.ndarrays of the rotated MTF positions in x and y directions
        """

        original_mtf_x_position = np.array([x * spacing[0] for x in range(20)])
        original_mtf_x_positions = copy.copy(original_mtf_x_position)
        for row in range(19):
            original_mtf_x_positions = np.row_stack(
                (original_mtf_x_positions, original_mtf_x_position)
            )

        original_mtf_y_position = np.array([x * spacing[1] for x in range(20)])
        original_mtf_y_positions = copy.copy(original_mtf_y_position)
        for row in range(19):
            original_mtf_y_positions = np.column_stack(
                (original_mtf_y_positions, original_mtf_y_position)
            )

        # we are only interested in the rotated y positions as there correspond to the distance of the data from the edge
        rotated_mtf_y_positions = -original_mtf_x_positions * np.sin(angle) + (
            original_mtf_y_positions - intercept
        ) * np.cos(angle)

        rotated_mtf_x_positions = original_mtf_x_positions * np.cos(angle) + (
            original_mtf_y_positions - intercept
        ) * np.sin(angle)

        return rotated_mtf_x_positions, rotated_mtf_y_positions

    def get_esf(self, edge_arr, y):
        """Extract the edge response function

        Args:
            edge_arr (np.ndarray): _description_
            y (np.ndarray): _description_

        Returns:
            tuple: u and esf - 'normal' and interpolated edge response function (ESF)
        """

        # extract the distance from the edge and the corresponding data as vectors

        edge_distance = copy.copy(y[0, :])

        for row in range(1, 20):
            edge_distance = np.append(edge_distance, y[row, :])

        esf_data = copy.copy(edge_arr[:, 0])
        for row in range(1, 20):
            esf_data = np.append(esf_data, edge_arr[:, row])

        # sort the distances and the data accordingly
        ind_edge_distance = np.argsort(edge_distance)
        sorted_edge_distance = edge_distance[ind_edge_distance]
        sorted_esf_data = esf_data[ind_edge_distance]

        # get rid of duplicates (if two data correspond to the same distance) and replace them with their average
        temp_array01 = np.array([sorted_edge_distance[0]])
        temp_array02 = np.array([sorted_esf_data[0]])

        for element in range(1, len(sorted_edge_distance)):
            if not (sorted_edge_distance[element] - temp_array01[-1]).all():
                temp_array02[-1] = (
                    temp_array02[-1] + sorted_esf_data[element]
                ) / 2

            else:
                temp_array01 = np.append(
                    temp_array01, sorted_edge_distance[element]
                )
                temp_array02 = np.append(
                    temp_array02, sorted_esf_data[element]
                )

        # ;interpolate the edge response function (ESF) so that it only has 128 elements
        u = np.linspace(temp_array01[0], temp_array01[-1], 128)
        esf = np.interp(u, temp_array01, temp_array02)

        return u, esf

    def calculate_mtf_for_edge(self, dcm, edge):
        pixels = dcm.pixel_array
        pe = get_pe_direction(dcm)

        img = rescale_to_byte(pixels)  # rescale for OpenCV operations
        thresh = self.thresh_image(img)
        circle = self.get_circles(img)
        circle_radius = circle[0][0][2]
        square, box = self.find_square(thresh)
        if edge == "right":
            _, centre = self.get_right_edge_vector_and_centre(square)
        else:
            _, centre = self.get_top_edge_vector_and_centre(square)

        edge_arr = self.get_edge_roi(pixels, centre)
        void_arr = self.get_void_roi(pixels, circle)
        signal_arr = self.get_signal_roi(pixels, edge, centre, circle_radius)
        spacing = get_pixel_size(dcm)
        mean = np.mean([void_arr, signal_arr])
        x_edge, y_edge, edge_arr = self.get_edge(edge_arr, mean, spacing)
        angle, intercept = self.get_edge_angle_and_intercept(x_edge, y_edge)
        x, y = self.get_edge_profile_coords(angle, intercept, spacing)
        u, esf = self.get_esf(edge_arr, y)
        # This function calculated the LSF by taking the derivative of the ESF.
        # Reference: https://www.ncbi.nlm.nih.gov/pmc/articles/PMC3643984/
        lsf = np.gradient(esf)
        lsf = np.array(lsf)
        n = lsf.size
        mtf = abs(np.fft.fft(lsf))
        norm_mtf = mtf / mtf[0]
        mtf_50 = min(
            [
                i
                for i in range(len(norm_mtf) - 1)
                if norm_mtf[i] >= 0.5 >= norm_mtf[i + 1]
            ]
        )
        profile_length = max(y.flatten()) - min(y.flatten())
        freqs = fftfreq(n, profile_length / n)
        mask = freqs >= 0
        mtf_frequency = 10.0 * mtf_50 / profile_length
        res = 10 / (2 * mtf_frequency)

        if self.report:
            import matplotlib.pyplot as plt

            fig, axes = plt.subplots(11, 1)
            fig.set_size_inches(5, 36)
            fig.tight_layout(pad=4)
            axes[0].set_title("raw pixels")
            axes[0].imshow(pixels, cmap="gray")
            axes[1].set_title("rescaled to byte")
            axes[1].imshow(img, cmap="gray")
            axes[2].set_title("thresholded")
            axes[2].imshow(thresh, cmap="gray")
            axes[3].set_title("finding circle")
            c = cv.circle(
                img,
                (circle[0][0][0], circle[0][0][1]),
                circle[0][0][2],
                (255, 0, 0),
            )
            axes[3].imshow(c)
            box = cv.drawContours(img, [box], 0, (255, 0, 0), 1)
            axes[4].set_title("finding MTF square")
            axes[4].imshow(box)
            axes[5].set_title("edge ROI")
            axes[5].imshow(edge_arr, cmap="gray")
            axes[6].set_title("void ROI")
            im = axes[6].imshow(void_arr, cmap="gray")
            fig.colorbar(im, ax=axes[6])
            axes[7].set_title("signal ROI")
            im = axes[7].imshow(signal_arr, cmap="gray")
            fig.colorbar(im, ax=axes[7])
            axes[8].set_title("edge spread function")
            axes[8].plot(esf)
            axes[8].set_xlabel("mm")
            axes[9].set_title("line spread function")
            axes[9].plot(lsf)
            axes[9].set_xlabel("mm")
            axes[10].set_title("normalised MTF")
            axes[10].plot(freqs[mask], norm_mtf[mask])
            axes[10].set_xlabel("lp/mm")
            logger.debug(
                f"Writing report image: {self.report_path}_{pe}_{edge}.png"
            )
            img_path = os.path.realpath(
                os.path.join(
                    self.report_path, f"{self.img_desc(dcm)}_{pe}_{edge}.png"
                )
            )
            fig.savefig(img_path)
            self.report_files.append(img_path)

        return res

    def calculate_mtf(self, dcm) -> tuple:
        pe = get_pe_direction(dcm)
        pe_result, fe_result = None, None

        if pe == "COL":
            pe_result = self.calculate_mtf_for_edge(dcm, "top")
            fe_result = self.calculate_mtf_for_edge(dcm, "right")
        elif pe == "ROW":
            pe_result = self.calculate_mtf_for_edge(dcm, "right")
            fe_result = self.calculate_mtf_for_edge(dcm, "top")

        return pe_result, fe_result



"""
ACR Uniformity

https://www.acraccreditation.org/-/media/acraccreditation/documents/mri/largephantomguidance.pdf

Calculates the percentage integral uniformity for slice 7 of the ACR phantom.

This script calculates the percentage integral uniformity in accordance with the ACR Guidance.
This is done by first defining a large 200cm2 ROI before placing 1cm2 ROIs at every pixel within
the large ROI. At each point, the mean of the 1cm2 ROI is calculated. The ROIs with the maximum and
minimum mean value are used to calculate the integral uniformity. The results are also visualised.

Created by Yassine Azma
yassine.azma@rmh.nhs.uk

13/01/2022
"""

import os
import sys
import traceback

import numpy as np
from hazenlib.ACRObject import ACRObject
from hazenlib.HazenTask import HazenTask
from hazenlib.logger import logger
from hazenlib.types import Measurement, Result
from hazenlib.utils import (
    compute_radius_from_area,
    create_circular_mask,
    create_circular_mean_kernel,
    create_circular_roi_at,
    detect_roi_center,
)
from matplotlib.pyplot import subplots as plt_subplots
from scipy.signal import convolve2d


class ACRUniformity(HazenTask):
    """Uniformity measurement class for DICOM images of the ACR phantom."""

    def __init__(self, **kwargs):
        if kwargs.pop("verbose", None) is not None:
            logger.warning(
                "verbose is not a supported argument for %s",
                type(self).__name__,
            )
        super().__init__(**kwargs)

        # Initialise ACR object
        self.ACR_obj = ACRObject(self.dcm_list)
        # Required pixel radius to produce ~200cm2 ROI
        # (this produces 25000 mm2??)
        self.r_large = compute_radius_from_area(200, self.ACR_obj.dx)
        # Required pixel radius to produce ~1cm2 ROI (produces 150mm2)
        self.r_small = compute_radius_from_area(1, self.ACR_obj.dx)
        # Kernel we can use to convolve on input array to obtain an ROI mean
        self.r_small_kernel = create_circular_mean_kernel(self.r_small)
        logger.info(
            f"Generated 2D circular kernel for target => \n{self.r_small_kernel}"
        )
        # Required pixel radius to produce ~200cm2 ROI - 1cm to ensure small rois live fully within large ROI
        self.r_large_filter = self.r_large - self.r_small
        # Offset used when adding labels to plots. They display 10 mm to the bottom and right of the ROI
        self.label_x_offset = np.ceil(np.divide(10, self.ACR_obj.dx))

    def run(self) -> Result:
        """Main function for performing uniformity measurement using slice 7 from the ACR phantom image set.

        Returns:
            dict: results are returned in a standardised dictionary structure specifying the task name, input DICOM
                Series Description + SeriesNumber + InstanceNumber, task measurement key-value pairs, optionally path
                to the generated images for visualisation
        """
        # Initialise results dictionary
        results = self.init_result_dict(desc=self.ACR_obj.acquisition_type())
        results.files = self.img_desc(self.ACR_obj.slice_stack[6])

        try:
            result = self.get_integral_uniformity(self.ACR_obj.slice_stack[6])
            results.add_measurement(
                Measurement(
                    name="Uniformity",
                    type="measured",
                    subtype="Integral uniformity",
                    unit="%",
                    value=round(result, 2),
                ),
            )
        except Exception as e:
            logger.exception(
                "Could not calculate the percent integral uniformity for %s"
                " because of : %s",
                self.img_desc(self.ACR_obj.slice_stack[6]),
                e,
            )
            traceback.print_exc(file=sys.stdout)

        # only return reports if requested
        if self.report:
            results.add_report_image(self.report_files)

        return results

    def write_report(self, img, centre, min_roi, max_roi, piu, dcm):
        (centre_x, centre_y) = centre

        # To keep convention with the rest of this module
        # we'll unpack these variables x, y
        # with the note that x = rows and y = columns
        # which is different to the usual convention
        # of x = columns and y = rows.
        x_max, y_max, max_value = max_roi
        x_min, y_min, min_value = min_roi

        fig, axes = plt_subplots(2, 1)
        fig.set_size_inches(8, 16)
        fig.tight_layout(pad=4)

        theta = np.linspace(0, 2 * np.pi, 360)

        axes[0].imshow(img)
        axes[0].scatter(centre_y, centre_x, c="red")
        axes[0].axis("off")
        axes[0].set_title("Centroid Location")

        axes[1].imshow(img)
        axes[1].scatter([x_max, x_min], [y_max, y_min], c="red", marker="x")
        axes[1].plot(
            self.r_small * np.sin(theta) + x_max,
            self.r_small * np.cos(theta) + y_max,
            c="yellow",
        )
        axes[1].annotate(
            "Min = " + str(np.round(min_value, 1)),
            [x_min + self.label_x_offset, y_min],
            c="white",
        )

        axes[1].plot(
            self.r_small * np.sin(theta) + x_min,
            self.r_small * np.cos(theta) + y_min,
            c="yellow",
        )
        axes[1].annotate(
            "Max = " + str(np.round(max_value, 1)),
            [x_max + self.label_x_offset, y_max],
            c="white",
        )
        axes[1].plot(
            self.r_large * np.cos(theta) + centre_x,
            self.r_large * np.sin(theta) + centre_y,
            c="black",
        )
        axes[1].axis("off")
        axes[1].set_title(
            "Percent Integral Uniformity = " + str(np.round(piu, 2)) + "%"
        )

        img_path = os.path.realpath(
            os.path.join(self.report_path, f"{self.img_desc(dcm)}.png")
        )
        fig.savefig(img_path)
        self.report_files.append(img_path)

    def calculate_uniformity(self, min_value, max_value):
        """Calculates the percent integral uniformity (PIU) using formula `PIU = 100 * (1 - (max - min) / (max + min))`.

        It is broken down into constituent elements to shift the math to numpy as much as possible to squeeze
        performance as much as possible.

        Args:
            min_value (int): minimum mean value.
            max_value (int): maximum mean value.

        Returns:
            float: value of integral uniformity.

        """
        subtraction = np.subtract(max_value, min_value)
        addition = np.add(max_value, min_value)
        division = np.divide(subtraction, addition)
        fraction = np.subtract(1, division)
        return np.multiply(100, fraction)

    def get_mean_roi_values(self, img, search_space_mask=None):
        """
        This method gets the mean small rois for the areas of minimum and maximum intensities.

        Below is an excerpt from the ACR Large Phantom Guidance

        For each series, the measurements are made according to the following procedure:

            #. Display slice location 7.
            #. Place a large, circular region-of-interest (ROI) on the image as shown in Figure 15.
                *. The area of the ROI depends on whether the large (200 cm2) or medium (160 cm2) phantom
                    was scanned (Table 4).
                *. This large ROI defines the boundary of the region in which the image uniformity is measured.
                *. Although the mean pixel intensity inside this ROI is not needed for the uniformity test, it is
                    used in the percent signal ghosting test (section 6.0), so it should be noted.
            #. Set the display window to its minimum, and lower the level until the entire area inside the large ROI is
            white.
                *. The goal now is to raise the level slowly until a small, roughly 1 cm2 region of dark pixels
                    develops inside the ROI. This is the region of lowest signal in the large ROI.
                *. Sometimes more than one region of dark pixels will appear. In that case, focus attention on the
                    largest dark region.
                *. In some cases, rather than having a well-defined dark region, one or more wide, poorly defined
                    dark areas or areas of mixed black and white pixels are apparent.
                *. In that case, make a visual estimate of the location of the darkest 1 cm2 portion of the largest
                    dark area should be made.
            #. Place a 1 cm2 circular ROI on the low-signal region identified in step 3.
                *. If measuring in the Medium phantom be sure that this ROI does not include any of the notch at
                the top of the phantom.
                *. Figures 16a and 17a show what typical Large and Medium phantom images look like at this
                    point.
                *. Record the mean pixel value for this 1 cm2 ROI. This is the measured low-signal value.
                *. If there is uncertainty about where to place the ROI because there is no single obviously darkest
                location, try several locations and select the one having the lowest mean pixel value.
            #. Raise the level until all but a small, roughly 1 cm2 region of white pixels remains inside the large ROI.
                *. This is the region of highest signal.
                *. Sometimes more than 1 region of white pixels will remain.
                    In that case, focus attention on the largest white region. It can happen that rather than having a
                    well-defined white region, one ends up with 1 or more diffuse areas of mixed black and white pixels.
                    In that case, make a best estimate of the location of the brightest 1 cm2 portion of the largest
                    bright area.
            #. Place a 1 cm2 circular ROI on the high-signal region identified in step 5.
                *. Record the average pixel value for this 1 cm2 ROI. This is the measured high-signal value.
                *. If there is uncertainty about where to place the ROI because there is no single obviously
                    brightest location, try several locations and select the one having the highest mean pixel value.

        Args:
            img (np.ma.MaskedArray): Large ROI pixel data.
            search_space_mask (np.ndarray): Mask delineating any restrictions on the search space used to look for the
                minima and maxima.

        Returns:
            float: value of integral uniformity.

        """
        # First, let's make sure we zero out the periphery of our large ROI so that the convolution can ignore it.
        mask = img.mask
        large_roi = img.copy()
        large_roi[mask] = 0
        # Convolve the large ROI with our kernel.
        # See https://docs.opencv.org/3.0-beta/doc/py_tutorials/py_imgproc/py_morphological_ops/py_morphological_ops.html
        mean_large_roi = convolve2d(
            large_roi, self.r_small_kernel, mode="same"
        )
        # We want to bias the valid centers to be fully within the large ROI.
        # Per the ACR example, it looks like it should be biased by one radius (or two radii in our case due to implementation).
        valid_search_mask = (
            mask if search_space_mask is None else search_space_mask
        )
        mean_large_roi = np.ma.masked_array(
            mean_large_roi, mask=valid_search_mask, fill_value=0
        )
        # Find mean (which is already averaged) and x and y coordinates of that point in the averaged data.
        # Do this for both the minimum and maximum
        min_mean = mean_large_roi.min()
        x_min, y_min = detect_roi_center(
            mean_large_roi, mean_large_roi.argmin()
        )
        max_mean = mean_large_roi.max()
        x_max, y_max = detect_roi_center(
            mean_large_roi, mean_large_roi.argmax()
        )
        return x_min, y_min, min_mean, x_max, y_max, max_mean

    def get_integral_uniformity(self, dcm):
        """Calculates the percent integral uniformity (PIU) of a DICOM pixel array. \n
        Iterates with a ~1 cm^2 ROI through a ~200 cm^2 ROI inside the phantom region,
        and calculates the mean non-zero pixel value inside each ~1 cm^2 ROI. \n
        The PIU is defined as: `PIU = 100 * (1 - (max - min) / (max + min))`, where \n
        'max' and 'min' represent the maximum and minimum of the mean non-zero pixel values of each ~1 cm^2 ROI.

        Args:
            dcm (pydicom.Dataset): DICOM image object to calculate uniformity from.

        Returns:
            float: value of integral uniformity.
        """
        img = dcm.pixel_array

        (center_x, center_y), _ = self.ACR_obj.find_phantom_center(
            img, self.ACR_obj.dx, self.ACR_obj.dy
        )
        logger.info(f"Adjusted centroid to ({center_x}, {center_y})")

        logger.info("Getting large ROI in image...")
        large_roi = create_circular_roi_at(
            img, self.r_large, center_x, center_y
        )
        large_roi_valid_space = create_circular_mask(
            img, self.r_large_filter, center_x, center_y
        )

        logger.info("Getting the min and max mean ROIs in image...")
        x_min, y_min, min_value, x_max, y_max, max_value = (
            self.get_mean_roi_values(large_roi, ~large_roi_valid_space)
        )
        logger.info(f"Mean Min ROI => ({x_min}, {y_min}) = {min_value}")
        logger.info(f"Mean Max ROI => ({x_max}, {y_max}) = {max_value}")

        # Uniformity calculation
        piu = self.calculate_uniformity(min_value, max_value)
        logger.info(
            f"PIU[{piu}] = 100 * (1 - ({max_value} - {min_value}) / ({max_value} + {min_value}))"
        )

        if self.report:
            logger.info("Writing report ... ")
            self.write_report(
                img,
                (center_x, center_y),
                (x_min, y_min, min_value),
                (x_max, y_max, max_value),
                piu,
                dcm,
            )

        return piu
